# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for the MapdlPlotter class."""
from collections import OrderedDict
from typing import Any, Iterable, Optional, Union

from ansys.tools.visualization_interface import Plotter
from ansys.tools.visualization_interface.backends.pyvista import PyVistaBackendInterface
import numpy as np
from numpy.typing import NDArray

from ansys.mapdl.core import _HAS_VISUALIZER
from ansys.mapdl.core.misc import get_bounding_box
from ansys.mapdl.core.plotting.consts import (
    ALLOWED_TARGETS,
    BC_D,
    BC_F,
    BCS,
    FIELDS,
    FIELDS_ORDERED_LABELS,
    POINT_SIZE,
)
from ansys.mapdl.core.plotting.theme import MapdlTheme

if _HAS_VISUALIZER:
    import pyvista as pv

    from ansys.mapdl.core.plotting.plotting_defaults import DefaultSymbol

    BC_plot_settings = DefaultSymbol()


class MapdlPlotterBackend(PyVistaBackendInterface):
    """Provides the plotter for PyMAPDL.

    This class is an implementation of the ``PlotterInterface`` class from the ``ansys-visualizer`` package.
    Picker is implemented in PyMAPDL-specific classes due to the characteristics of the library.

    Parameters
    ----------
    use_trame : bool, optional
        Whether to use the Trame interface or not. The default is ``False``.
    theme : pv.DefaultTheme, optional
        Theme to use for the plotter. The default is ``None``.
    """

    def __init__(
        self, use_trame: bool = False, plot_picked_names: bool = True, **plotter_kwargs
    ):
        """Initialize the ``MapdlPlotter`` class."""
        super().__init__(
            use_trame=use_trame, plot_picked_names=plot_picked_names, **plotter_kwargs
        )

    def plot_iter(
        self, plottable_object: Any, name_filter: str = None, **plotting_options
    ):
        pass

    def plot(self, plottable_object: Any, name_filter: str = None, **plotting_options):
        pass

    @property
    def scene(self):
        """Return the scene."""
        return self.pv_interface.scene

    @scene.setter
    def scene(self, scene):
        """Set the scene."""
        self.pv_interface._scene = scene

    @property
    def meshes(self):
        """Return the meshes."""
        return self.scene.meshes


class MapdlPlotter(Plotter):
    """Plotter class for PyMAPDL.

    This class is an implementation of the ``PlotterInterface`` class from the `Visualization Interface Tool <https://github.com/ansys/ansys-tools-visualization-interface>`_.
    The picker is implemented in PyMAPDL-specific classes due to the characteristics of the library.

    Parameters
    ----------
    use_trame : bool, optional
        Whether to use the Trame visualizer. The default is ``False``.
    theme : pv.DefaultTheme, optional
        Theme to use for the plotter. The default is ``None``.
    """

    def __init__(
        self, use_trame: bool = False, theme: pv.Plotter.theme = None, **plotter_kwargs
    ):
        """Initialize the ``MapdlPlotter`` class."""
        self._backend = MapdlPlotterBackend(use_trame=use_trame, **plotter_kwargs)
        super().__init__(backend=self._backend)
        self._theme = theme
        if theme is None:
            self._theme = MapdlTheme()
        self._off_screen = None
        self._notebook = None
        self._savefig = None
        self._title = None

    def _bc_labels_checker(self, bc_labels):
        """Make sure we have allowed parameters and data types for ``bc_labels``"""
        if not isinstance(bc_labels, (str, list, tuple)):
            raise ValueError(
                "The parameter 'bc_labels' can be only a string, a list of strings or tuple of strings."
            )

        if isinstance(bc_labels, str):
            bc_labels = bc_labels.upper()
            if bc_labels not in BCS and bc_labels not in FIELDS:
                raise ValueError(
                    f"The parameter '{bc_labels}' in 'bc_labels' is not supported.\n"
                    "Please use any of the following:\n"
                    f"{FIELDS}\nor any combination of:\n{BCS}"
                )
            bc_labels = [bc_labels]

        elif isinstance(bc_labels, (list, tuple)):
            if not all([isinstance(each, str) for each in bc_labels]):
                raise ValueError(
                    "The parameter 'bc_labels' can be a list or tuple, but it should only contain strings inside them."
                )

            if not all([each.upper() in BCS for each in bc_labels]):
                raise ValueError(
                    f"One or more parameters in 'bc_labels'({bc_labels}) are not supported.\n"
                    f"Please use any combination of the following:\n{BCS}"
                )

            bc_labels = [
                each.upper() for each in set(bc_labels)
            ]  # Removing duplicates too

        # Replacing field by equivalent fields.
        for each_field in FIELDS:
            if each_field in bc_labels:
                bc_labels.remove(each_field)
                bc_labels.extend(FIELDS[each_field])

        return bc_labels

    def _bc_target_checker(self, bc_target):
        """Make sure we have allowed parameters and data types for ``bc_labels``"""
        if not isinstance(bc_target, (str, list, tuple)):
            raise ValueError(
                "The parameter 'bc_target' can be only a string, a list of strings or tuple of strings."
            )

        if isinstance(bc_target, str):
            if bc_target.upper() not in ALLOWED_TARGETS:
                raise ValueError(
                    f"The parameter '{bc_target}' in 'bc_target' is not supported.\n"
                    f"At the moments only the following are supported:\n{ALLOWED_TARGETS}"
                )
            bc_target = [bc_target.upper()]

        if isinstance(bc_target, (list, tuple)):
            if not all([each.upper() in ALLOWED_TARGETS for each in bc_target]):
                raise ValueError(
                    "One or more parameters in 'bc_target' are not supported.\n"
                    f"Please use any combination of the following:\n{ALLOWED_TARGETS}"
                )
            if not all([isinstance(each, str) for each in bc_target]):
                raise ValueError(
                    "The parameter 'bc_target' can be a list or tuple, but it should only contain strings inside them."
                )

            bc_target = [each.upper() for each in bc_target]

        return bc_target

    def _bc_labels_default(self, mapdl):
        """Get the labels from the MAPDL database."""
        flist = list(
            set(
                [
                    each[1].upper()
                    for each in mapdl.get_nodal_loads()
                    if each[1].upper() in BCS
                ]
            )
        )
        dlist = list(
            set(
                [
                    each[1].upper()
                    for each in mapdl.get_nodal_constrains()
                    if each[1].upper() in BCS
                ]
            )
        )
        return flist + dlist

    def get_meshes_from_plotter(self):
        """Get the meshes plotted in the plotter.

        Returns
        -------
        list[pv.PolyData]
            Plotted meshes.
        """
        datasets = []
        for actor in self.scene.actors.values():
            if hasattr(actor, "mapper"):
                datasets.append(actor.mapper.dataset)

        return [
            actor.mapper.dataset
            for actor in self.scene.actors.values()
            if hasattr(actor, "mapper")
        ]

    def add_labels(
        self, points: Iterable[float], labels: Iterable[str], **plotting_options
    ) -> None:
        """Add labels to the plotter.

        Parameters
        ----------
        points : List[float]
            List of points for the labels.
        labels : List[str]
            List of labels to add.
        """
        self.scene.add_point_labels(points, labels, **plotting_options)

    def add_points(self, points: Iterable[float], **plotting_options) -> None:
        """Add points to the plotter.

        Parameters
        ----------
        points : List[float]
            List of points to add to the plotter.
        """
        _ = self.scene.add_points(points, **plotting_options)

    def plot_iter(
        self,
        plotting_list: Iterable[Any],
        name_filter: str = None,
        **plotting_options,
    ) -> None:
        """Add a list of objects to the plotter.

        Parameters
        ----------
        plotting_list : Iterable[Any]
            Iterable of objects to add to the plotter.
        name_filter : str, optional
            Filter to apply to the objects. The default is ``None``.
        """
        for plottable_object in plotting_list:
            self.plot(plottable_object, name_filter, **plotting_options)

    def add_mesh(
        self,
        meshes,
        points,
        labels,
        *,
        cpos=None,
        show_bounds=False,
        show_axes=True,
        background=None,
        # add_mesh kwargs:
        style=None,
        color="w",
        show_edges=None,
        edge_color=None,
        point_size=POINT_SIZE,
        line_width=None,
        opacity=1.0,
        flip_scalars=False,
        lighting=None,
        n_colors=256,
        interpolate_before_map=True,
        cmap=None,
        render_points_as_spheres=False,
        render_lines_as_tubes=False,
        scalar_bar_args=None,
        smooth_shading=None,
        feature_angle=30.0,
        show_scalar_bar=None,
        split_sharp_edges=None,
        # labels kwargs
        font_size=None,
        font_family=None,
        text_color=None,
        theme=None,
        add_points_kwargs=None,
        add_mesh_kwargs=None,
        add_point_labels_kwargs=None,
        plotter_kwargs=None,
    ):
        """Adds a mesh to the plotter.

        Parameters
        ----------
        meshes : pv.PolyData
            The mesh to add to the plotter.
        points : pv.PolyData
            The points to add to the plotter.
        labels : pv.PolyData
            The labels to add to the plotter.
        plotter_kwargs : dict, optional
            Extra kwargs, by default {}
        """
        if theme is None:
            theme = MapdlTheme()

        if not cmap:
            cmap = "bwr"

        if background:
            self.scene.set_background(background)
        else:
            self.scene.set_background("paraview")

        # Making sure that labels are visible in dark backgrounds
        if not text_color and background:
            bg = self.scene.background_color.float_rgb
            # from: https://graphicdesign.stackexchange.com/a/77747/113009
            gamma = 2.2
            threshold = (
                0.2126 * bg[0] ** gamma
                + 0.7152 * bg[1] ** gamma
                + 0.0722 * bg[2] ** gamma
                > 0.5 * gamma
            )
            if threshold:
                text_color = "black"
            else:
                text_color = "white"

        for point in points:
            self.add_points(
                point["points"],
                # scalars=point.get("scalars", None),
                color=color,
                show_edges=show_edges,
                edge_color=edge_color,
                point_size=point_size,
                line_width=line_width,
                opacity=opacity,
                flip_scalars=flip_scalars,
                lighting=lighting,
                n_colors=n_colors,
                interpolate_before_map=interpolate_before_map,
                cmap=cmap,
                render_points_as_spheres=render_points_as_spheres,
                render_lines_as_tubes=render_lines_as_tubes,
                **(add_points_kwargs or {}),
            )

        for mesh in meshes:
            scalars: Optional[NDArray[Any]] = mesh.get("scalars")

            if (
                "scalars" in mesh
                and scalars.ndim == 2
                and (scalars.shape[1] == 3 or scalars.shape[1] == 4)
            ):
                # for the case we are using scalars for plotting
                rgb = True
            else:
                rgb = False

            # To avoid index error.
            mesh_ = mesh["mesh"]
            if not isinstance(mesh_, list):
                mesh_ = [mesh_]

            for each_mesh in mesh_:
                self.scene.add_mesh(
                    each_mesh,
                    scalars=scalars,
                    scalar_bar_args=scalar_bar_args,
                    color=mesh.get("color", color),
                    style=mesh.get("style", style),
                    show_edges=show_edges,
                    edge_color=edge_color,
                    smooth_shading=smooth_shading,
                    split_sharp_edges=split_sharp_edges,
                    feature_angle=feature_angle,
                    point_size=point_size,
                    line_width=line_width,
                    show_scalar_bar=show_scalar_bar,
                    opacity=opacity,
                    flip_scalars=flip_scalars,
                    lighting=lighting,
                    n_colors=n_colors,
                    interpolate_before_map=interpolate_before_map,
                    cmap=cmap,
                    render_points_as_spheres=render_points_as_spheres,
                    render_lines_as_tubes=render_lines_as_tubes,
                    rgb=rgb,
                    **(add_mesh_kwargs or {}),
                )

        for label in labels:
            # Pyvista does not support plotting two labels with the same point.
            # It does handle that case by keeping only the first label.

            self.add_labels(
                label["points"],
                label["labels"],
                show_points=False,
                shadow=False,
                font_size=font_size,
                font_family=font_family,
                text_color=text_color,
                always_visible=True,
                **(add_point_labels_kwargs or {}),
            )

        if cpos:
            self.scene.camera_position = cpos

        if show_bounds:
            self.scene.show_bounds()

    def bc_plot(
        self,
        mapdl=None,
        bc_labels=None,
        bc_target=None,
        plot_bc_labels=False,
        plot_bc_legend=None,
        bc_glyph_size=None,
        bc_labels_font_size=16,
    ):
        """Plot boundary conditions.

        Parameters
        ----------
        mapdl : Mapdl, optional
            Mapdl instance, by default None
        bc_labels : optional
            Labels to plot, by default None
        bc_target : optional
            Targets to plot, by default None
        plot_bc_labels : bool, optional
            Whether to plot the labels or not, by default False
        plot_bc_legend : optional
            Legend to plot, by default None
        bc_glyph_size : optional
            Size of the gliph, by default None
        bc_labels_font_size : int, optional
            Size of the labels font, by default 16
        """
        if bc_labels:
            bc_labels = self._bc_labels_checker(bc_labels)
        else:
            bc_labels = self._bc_labels_default(mapdl)

        if bc_target:
            bc_target = self._bc_target_checker(bc_target)
        else:
            bc_target = ["NODES"]  # bc_target_default()

        # We need to scale the glyphs, otherwise, they will be plot sized 1.
        # We are going to calculate the distance between the closest points,
        # so we will scaled to a percentage of this distance.
        # This might create very small points in cases there are a concentration of points.
        #
        # Later can find a way to plot them and keep their size constant independent of the zoom.

        ratio = 0.075  # Because a glyph of 1 is too big.
        if bc_glyph_size is None:
            bc_glyph_size = get_bounding_box(mapdl.mesh.nodes)
            bc_glyph_size = bc_glyph_size[bc_glyph_size != 0]

            if bc_glyph_size.size != 0:
                bc_glyph_size = bc_glyph_size.mean() * ratio
            else:  # Case were there is only one node
                bc_glyph_size = ratio

        if not isinstance(bc_glyph_size, (int, float)):
            raise ValueError(
                "The 'bc_glyph_size' parameter can be only an int or float."
            )

        if "NODES" in bc_target:
            self.bc_nodes_plot(
                mapdl=mapdl,
                bc_labels=bc_labels,
                plot_bc_labels=plot_bc_labels,
                bc_glyph_size=bc_glyph_size,
                plot_bc_legend=plot_bc_legend,
                bc_labels_font_size=bc_labels_font_size,
            )

        # Add next things
        if "ELEM" in bc_target:
            pass

    def bc_nodes_plot(
        self,
        mapdl,
        bc_labels,
        plot_bc_labels=False,
        bc_glyph_size=1,
        plot_bc_legend=None,
        bc_labels_font_size=16,
    ):
        """Plot nodes BC given a list of labels."""
        nodes_xyz = mapdl.mesh.nodes
        nodes_num = mapdl.mesh.nnum

        bc_point_labels = None

        for each_label in bc_labels:
            if each_label in BC_D:
                bc = mapdl.get_nodal_constrains(each_label)

            elif each_label in BC_F:
                bc = mapdl.get_nodal_loads(each_label)

            else:
                raise Exception(f"The label '{each_label}' is not supported.")

            if bc.size == 0:  # There is no nodes with such label
                return

            bc_num = bc[:, 0].astype(int)
            bc_nodes = nodes_xyz[np.isin(nodes_num, bc_num), :]
            bc_values = bc[:, 1:].astype(float)
            bc_scale = abs(bc_values[:, 0])

            if bc_scale.max() != bc_scale.min():
                bc_scale = (bc_scale - bc_scale.min()) / (
                    bc_scale.max() - bc_scale.min()
                ) + 0.5  # Normalization around 0.5
            else:
                bc_scale = (
                    np.ones(bc_scale.shape) * 0.5
                )  # In case all the values are the same

            pcloud = pv.PolyData(bc_nodes)
            pcloud["scale"] = bc_scale * bc_glyph_size

            # Specify tolerance in terms of fraction of bounding box length.
            # Float value is between 0 and 1. Default is None.
            # If absolute is True then the tolerance can be an absolute distance.
            # If None, points merging as a preprocessing step is disabled.
            glyphs = pcloud.glyph(
                orient=False,
                scale="scale",
                # tolerance=0.05,
                geom=BC_plot_settings(each_label)["glyph"],
            )
            name_ = f"{each_label}"
            self.scene.add_mesh(
                glyphs,
                # name_filter=None,
                color=BC_plot_settings(each_label)["color"],
                style="surface",
                # style='wireframe',
                # line_width=3,
                name=name_,
                label=name_,
                opacity=0.40,
            )

            if plot_bc_labels:
                if bc_point_labels is None:
                    bc_point_labels = {each: "" for each in nodes_num}

                for id_, values in zip(bc_num, bc_values):
                    if not bc_point_labels[id_]:
                        bc_point_labels[id_] = (
                            f"Node: {id_}\n{each_label}: {values[0]:6.3f}, {values[1]:6.3f}"
                        )
                    else:
                        bc_point_labels[id_] = (
                            f"{bc_point_labels[id_]}\n{each_label}: {values[0]:6.3f}, {values[1]:6.3f}"
                        )

        if plot_bc_labels:
            pcloud = pv.PolyData(nodes_xyz)
            pcloud["labels"] = list(bc_point_labels.values())
            self.add_labels(
                pcloud.points,
                pcloud["labels"],
                shape_opacity=0.25,
                font_size=bc_labels_font_size,
                # There is a conflict here. See
                # To do not hide the labels, even when the underlying nodes
                # are hidden, we set "always_visible"
                always_visible=True,
                show_points=False,  # to not have node duplicity
                font_family="arial",
            )

        if plot_bc_legend:
            # Reorder labels to keep a consistent order
            sorted_dict = OrderedDict()
            labels_ = self.scene.renderer._labels.copy()

            # sorting the keys
            for symbol in FIELDS_ORDERED_LABELS:
                for key, value in labels_.items():
                    # taking advantage and overriding the legend glyph with
                    # something it can be seen properly in the legend
                    label_ = value[1]
                    if "U" in label_:
                        value = [BC_plot_settings("UY")["glyph"], label_, value[2]]
                    elif "F" in label_:
                        value = [BC_plot_settings("FX")["glyph"], label_, value[2]]
                    else:
                        value = [BC_plot_settings(label_)["glyph"], label_, value[2]]

                    if symbol == value[1]:
                        sorted_dict[key] = value

            # moving the not added labels (just in case)
            for key, value in labels_.items():
                if label_ not in FIELDS_ORDERED_LABELS:
                    sorted_dict[key] = value

            # overwriting labels
            self.scene.renderer._labels = sorted_dict
            self.scene.theme.font.family = "arial"
            self.scene.add_legend(bcolor=None)

    def plot(
        self,
        meshes,
        points,
        labels,
        *,
        title="",
        cpos=None,
        show_bounds=False,
        show_axes=True,
        background=None,
        # add_mesh kwargs:
        style=None,
        color="w",
        show_edges=None,
        edge_color=None,
        point_size=5.0,
        line_width=None,
        opacity=1.0,
        flip_scalars=False,
        lighting=None,
        n_colors=256,
        interpolate_before_map=True,
        cmap=None,
        render_points_as_spheres=False,
        render_lines_as_tubes=False,
        scalar_bar_args=None,
        smooth_shading=None,
        show_scalar_bar=None,
        split_sharp_edges=None,
        # labels kwargs
        font_size=None,
        font_family=None,
        text_color=None,
        theme=None,
        return_plotter=False,
        return_cpos=False,
        mapdl=None,
        plot_bc=False,
        plot_bc_legend=None,
        plot_bc_labels=None,
        bc_labels=None,
        bc_target=None,
        bc_glyph_size=None,
        bc_labels_font_size=16,
        add_points_kwargs=None,
        add_mesh_kwargs=None,
        add_point_labels_kwargs=None,
        plotter_kwargs=None,
        **kwargs,
    ) -> None:
        """Add an object to the plotter.

        Parameters
        ----------
        object : Any
            Object to add to the plotter.
        name_filter : str, optional
            Filter to apply to the object. The default is ``None``.
        """

        # Getting the plotter
        self.add_mesh(
            meshes,
            points,
            labels,
            cpos=cpos,
            show_bounds=show_bounds,
            show_axes=show_axes,
            background=background,
            # add_mesh kwargs:
            style=style,
            color=color,
            show_edges=show_edges,
            edge_color=edge_color,
            point_size=point_size,
            line_width=line_width,
            opacity=opacity,
            flip_scalars=flip_scalars,
            lighting=lighting,
            n_colors=n_colors,
            interpolate_before_map=interpolate_before_map,
            cmap=cmap,
            render_points_as_spheres=render_points_as_spheres,
            render_lines_as_tubes=render_lines_as_tubes,
            scalar_bar_args=scalar_bar_args,
            smooth_shading=smooth_shading,
            show_scalar_bar=show_scalar_bar,
            split_sharp_edges=split_sharp_edges,
            # labels kwargs
            font_size=font_size,
            font_family=font_family,
            text_color=text_color,
            theme=theme,
            add_points_kwargs=add_points_kwargs,
            add_mesh_kwargs=add_mesh_kwargs,
            add_point_labels_kwargs=add_point_labels_kwargs,
            plotter_kwargs=plotter_kwargs,
        )

        if plot_bc:
            if not mapdl:
                raise ValueError(
                    "An instance of `ansys.mapdl.core.mapdl.MapdlBase` "
                    "should be passed using `mapdl` keyword if you are aiming "
                    "to plot the boundary conditions (`plot_bc` is `True`)."
                )
            self.bc_plot(
                mapdl=mapdl,
                plot_bc_legend=plot_bc_legend,
                plot_bc_labels=plot_bc_labels,
                bc_labels=bc_labels,
                bc_target=bc_target,
                bc_glyph_size=bc_glyph_size,
                bc_labels_font_size=bc_labels_font_size,
            )

        if title:  # Added here to avoid labels overlapping title
            self.scene.add_title(title, color=text_color)

        if return_cpos and return_plotter:
            raise ValueError(
                "'return_cpos' and 'return_plotter' cannot be both 'True' at the same time."
            )

    def switch_scene(self, pl: Union["pv.Plotter", "MapdlPlotter"]) -> "MapdlPlotter":
        """Switches the backend scene to the given plotter.

        Parameters
        ----------
        pl : Union["pv.Plotter", "MapdlPlotter"]
            Plotter to change the scene to.

        Returns
        -------
        MapdlPlotter
            The MapdlPlotter instance.
        """
        if isinstance(pl, pv.Plotter):
            self._backend.scene = pl
            self._backend.enable_widgets()
            return self
        elif isinstance(pl, MapdlPlotter):
            return pl
        elif pl is None:
            return self
        else:
            raise TypeError(
                f"Expected a ``MapdlPlotter`` or ``pv.Plotter`` instance, but got {type(pl)}"
            )

    def show(
        self,
        window_size=None,
        return_plotter=False,
        return_cpos=False,
        notebook=None,
        savefig=None,
        off_screen=None,
        **kwargs,
    ) -> None:
        """Show the plotter."""

        self._off_screen = off_screen
        if notebook:
            self._off_screen = True  # pragma: no cover

        if savefig:
            self._off_screen = True
            self._notebook = False
        # permit user to save the figure as a screenshot
        if self._savefig or savefig:
            self._backend.show(
                auto_close=False,
                window_size=window_size,
                screenshot=savefig,
                **kwargs,
            )
            self.scene.screenshot(self._savefig)

            # return unclosed plotter
            if return_plotter:
                return self

            # ifplotter.scene.set_background("paraview") not returning plotter, close right away
            self.scene.close()

        else:
            if not return_plotter:
                self._backend.show()

        if return_plotter:
            return self
        elif return_cpos:
            return self.scene.camera_position
        else:
            return None

    @property
    def scene(self):
        """Return the scene."""
        return self._backend.scene

    @property
    def meshes(self):
        """Return the meshes."""
        return self.scene.meshes



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Functions to download sample datasets from the pyansys data repository.
"""
from functools import wraps
import os
import shutil
import urllib.request
import zipfile

try:
    import requests

    _HAS_REQUESTS = True
except ModuleNotFoundError:
    _HAS_REQUESTS = False

from ansys.mapdl import core as pymapdl


def check_directory_exist(directory):
    # Wrapping LISTING FUNCTIONS.
    def wrap_function(func):
        @wraps(func)
        def inner_wrapper(*args, **kwargs):
            # Check if folder exists
            if not os.path.exists(directory):
                os.makedirs(directory)

            return func(*args, **kwargs)

        return inner_wrapper

    return wrap_function


def get_ext(filename):
    """Extract the extension of the filename"""
    ext = os.path.splitext(filename)[1].lower()
    return ext


def delete_downloads():
    """Delete all downloaded examples to free space or update the files"""
    if os.path.exists(pymapdl.EXAMPLES_PATH):
        shutil.rmtree(pymapdl.EXAMPLES_PATH)
    return True


@check_directory_exist(pymapdl.EXAMPLES_PATH)
def _decompress(filename):
    zip_ref = zipfile.ZipFile(filename, "r")
    zip_ref.extractall(pymapdl.EXAMPLES_PATH)
    return zip_ref.close()


def _get_file_url(filename, directory=None):
    if directory:
        return (
            f"https://github.com/ansys/example-data/raw/master/{directory}/{filename}"
        )
    return f"https://github.com/ansys/example-data/raw/master/{filename}"


def _check_url_exist(url):
    if not _HAS_REQUESTS:
        raise ModuleNotFoundError("Examples module requires request module")

    response = requests.get(url)
    if response.status_code == 200:
        return [True]
    else:
        return [False]


@check_directory_exist(pymapdl.EXAMPLES_PATH)
def _retrieve_file(url, filename, _test=False):
    # scape test
    if pymapdl.RUNNING_TESTS:
        return _check_url_exist(url)

    # First check if file has already been downloaded
    local_path = os.path.join(pymapdl.EXAMPLES_PATH, os.path.basename(filename))
    local_path_no_zip = local_path.replace(".zip", "")
    if os.path.isfile(local_path_no_zip) or os.path.isdir(local_path_no_zip):
        return local_path_no_zip, None

    # Perform download
    saved_file, resp = urllib.request.urlretrieve(url)
    shutil.move(saved_file, local_path)
    if get_ext(local_path) in [".zip"]:
        _decompress(local_path)
        local_path = local_path[:-4]
    return local_path, resp


def _download_file(filename, directory=None, _test=False):
    url = _get_file_url(filename, directory)
    try:
        return _retrieve_file(url, filename, _test)
    except Exception as e:  # Genering exception
        raise RuntimeError(
            "For the reason mentioned below, retrieving the file from internet failed.\n"
            "You can download this file from:\n"
            f"{url}\n"
            "\n"
            "The reported error message is:\n"
            f"{str(e)}"
        )


def download_bracket():
    """Download an IGS bracket.

    Examples
    --------
    >>> from ansys.mapdl.core import examples
    >>> filename = examples.download_bracket()
    >>> filename
    '/home/user/.local/share/ansys_mapdl_core/examples/bracket.iges'

    """
    return _download_file("bracket.iges", "geometry")[0]


def download_tech_demo_data(example, filename):
    """Download Tech Demos external data."""
    example = "tech_demos/" + example
    return _download_file(filename=filename, directory=example)[0]


def download_vtk_rotor():
    """Download rotor vtk file."""
    return _download_file("rotor.vtk", "geometry")[0]


def _download_rotor_tech_demo_vtk():
    """Download the rotor surface VTK file."""
    return _download_file("rotor2.vtk", "geometry")[0]


def download_example_data(filename, directory=None):
    return _download_file(filename, directory=directory)[0]


def download_manifold_example_data() -> dict:
    """Download the manifold example data and return the
    download paths into a dictionary domain id->path.
    Examples files are downloaded to a persistent cache to avoid
    re-downloading the same file twice.
    Returns
    -------
    dict[str:str]
        Path to the example files.
    Examples
    --------
    Download the manifold geometry, ans file and return the path of the file
    >>> from ansys.mapdl.core.examples.downloads import download_manifold_example_data
    >>> paths = download_manifold_example_data()
    >>> paths
    {geometry: 'C:\\Users\\user\\AppData\\Local\\ansys_mapdl_core\\ansys_mapdl_core\\examples\\manifold_geometry.anf',
     mapping_data: 'C:\\Users\\user\\AppData\\Local\\ansys_mapdl_core\\ansys_mapdl_core\\examples\\manifold_cht-final_temp.csv'}
    """

    files_dir = "pymapdl/manifold"
    return {
        "geometry": _download_file(
            filename="manifold_geometry.anf", directory=files_dir
        )[0],
        "mapping_data": _download_file(
            filename="manifold_cht-final_temp.csv", directory=files_dir
        )[0],
    }


def download_cfx_mapping_example_data() -> dict:
    """Download the CFX mapping data and return the
    download paths into a dictionary domain id->path.
    Examples files are downloaded to a persistent cache to avoid
    re-downloading the same file twice.
    Returns
    -------
    dict[str:str]
        Path to the example files.
    Examples
    --------
    >>> from ansys.mapdl.core.examples.downloads import download_cfx_mapping_example_data
    >>> paths = download_cfx_mapping_example_data()
    >>> paths
    {data: 'C:\\Users\\user\\AppData\\Local\\ansys_mapdl_core\\ansys_mapdl_core\\examples\\11_blades_mode_1_ND_0.csv',
     model: 'C:\\Users\\user\\AppData\\Local\\ansys_mapdl_core\\ansys_mapdl_core\\examples\\ExampleMapping.db'}
    """

    files_dir = "pymapdl/cfx_mapping"
    return {
        "data": _download_file(
            filename="11_blades_mode_1_ND_0.csv", directory=files_dir
        )[0],
        "model": _download_file(filename="ExampleMapping.db", directory=files_dir)[0],
    }



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional

from ansys.mapdl.core.mapdl_types import MapdlInt


class Setup:
    def ansol(
        self,
        nvar="",
        node="",
        item="",
        comp="",
        name="",
        mat="",
        real="",
        ename="",
        **kwargs,
    ):
        """Specifies averaged nodal data to be stored from the results file in the

        APDL Command: ANSOL
        solution coordinate system.

        Parameters
        ----------
        nvar
            Arbitrary reference number assigned to this variable (2 to NV
            [NUMVAR]). Overwrites any existing results for this variable.

        node
            Node number for which data are to be stored.

        item
            Label identifying the item. General item labels are shown in
            Table 126: ANSOL - General Item and Component Labels below. Some
            items also require a component label.

        comp
            Component of the item (if required). General component labels are
            shown in Table 126: ANSOL - General Item and Component Labels
            below.

        name
            Thirty-two character name for identifying the item on the printout
            and displays. Defaults to an eight character label formed by
            concatenating the first four characters of the Item and Comp
            labels.

        mat
            The material number. Average will be computed based on the subset
            of elements with the specified material number. DEFAULT: Use all
            elements in the active set unless Real and/or Ename is specified.

        real
            The real number. Average will be computed based on the subset of
            elements with the specified real number. DEFAULT: Use all elements
            in the active set unless Mat and/or Ename is specified.

        ename
            The element type name. Average will be computed based on the subset
            of elements with the specified element type name. DEFAULT: Use all
            elements in the active set unless Mat and/or Real is specified.

        Notes
        -----
        Valid item and component labels for averaged nodal results are listed
        in Table: 126:: ANSOL - General Item and Component Labels, below.

        All element nodal quantities are obtained in RSYS, Solu and then
        averaged.

        The ANSOL command defines averaged nodal results data to be stored from
        a results file [FILE]. Not all items are valid for all nodes. See the
        input and output summary tables of the Element Reference of each
        element that is attached to the node for the available items.

        COORDINATE SYSTEMS: All element nodal results used by ANSOL for
        averaging are in the element coordinate system, except for layered
        elements. Layered element results are in the layer coordinate system.
        You can further specify the element nodal results, for some elements,
        with the SHELL, LAYERP26, and FORCE commands.

        ANSOL does not transform results from RSYS, SOLU to other coordinate
        systems. Verify that all elements attached to the subject node have the
        same coordinate system before using ANSOL.

        SHELL ELEMENTS: The default shell element coordinate system is based on
        node ordering. For shell elements the adjacent elements could have a
        different RSYS,SOLU, making the resultant averaged data inconsistent. A
        note to this effect is issued when ANSOL is used in models containing
        shell elements. Ensure that consistent coordinate systems are active
        for all associated elements used by the ANSOL command.

        DERIVED QUANTITIES: Some of the result items supported by ANSOL (see
        Table: 126:: ANSOL - General Item and Component Labels) are derived
        from the component quantities. Use AVPRIN to specify the principal and
        vector sum quantity averaging methods.

        DEFAULT: If Mat, Real , and Ename are not specified, all of the
        elements attached to the node will be considered. When a material ID,
        real constant ID, or element type discontinuity is detected at a node,
        a note is issued. For example, in a FSI analysis, a FLUID30 element at
        the structure interface would be considered. But since it contains no
        SX result, it will not be used during STORE operations.

        Table: 126:: : ANSOL - General Item and Component Labels

        For more information on the meaning of contact status and its possible
        values, see Reviewing Results in POST1 in the Contact Technology Guide.
        """
        command = f"ANSOL,{nvar},{node},{item},{comp},{name},{mat},{real},{ename}"
        return self.run(command, **kwargs)

    def cisol(self, n="", id_="", node="", cont="", dtype="", **kwargs):
        """Stores fracture parameter information in a variable.

        APDL Command: CISOL

        Parameters
        ----------
        n
            Arbitrary reference number or name assigned to this variable.
            Number must be >1 but </= NUMVAR.

        id\\_
            Crack ID number.

        node
            Crack tip node number.

        cont
            Contour number.

        dtype
            Data type to output:

            JINT - J-integral

            IIN1 - Interaction integral 1

            IIN2  - Interaction integral 2

            IIN3 - Interaction integral 3

            K1 - Mode 1 stress-intensity factor

            K2 - Mode 2 stress-intensity factor

            K3 - Mode 3 stress-intensity factor

            G1 - Mode 1 energy release rate

            G2  - Mode 2 energy release rate

            G3  - Mode 3 energy release rate

            GT - Total energy release rate

            MFTX - Total material force X

            MFTY - Total material force Y

            MFTZ - Total material force Z

            CEXT - Crack extension
        """
        command = f"CISOL,{n},{id_},{node},{cont},{dtype}"
        return self.run(command, **kwargs)

    def data(self, ir="", lstrt="", lstop="", linc="", name="", kcplx="", **kwargs):
        """Reads data records from a file into a variable.

        APDL Command: DATA

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting variable (2 to
            NV [NUMVAR]).  If this number is the same as for a previously
            defined variable, the previously defined variable will be
            overwritten with this result.

        lstrt
            Start at location LSTRT (defaults to 1).

        lstop
            Stop at location LSTOP (defaults to LSTRT).  Maximum location
            available is determined from data previously stored.

        linc
            Fill every LINC location between LSTRT and LSTOP (defaults to 1).

        name
            Eight character name for identifying the variable on the printout
            and displays.  Embedded blanks are compressed upon output.

        kcplx
            Complex number key:

            0 - Data stored as the real part of the complex number.

            1 - Data stored as the imaginary part of the complex number.

        Notes
        -----
        This command must be followed by a format statement (on the next line)
        and the subsequent data records, and all must be on the same file (that
        may then be read with the /INPUT command).  The format specifies the
        number of fields to be read per record, the field width, and the
        placement of the decimal point (if one is not included in the data
        value).  The read operation follows the available FORTRAN FORMAT
        conventions of the system.  See the system FORTRAN manual for details.
        Any standard FORTRAN real format (such as (4F6.0), (F2.0,2X,F12.0),
        etc.) may be used.  Integer (I), character (A), and list-directed (*)
        descriptors may not be used.  The parentheses must be included in the
        format.  Up to 80 columns per record may be read.  Locations may be
        filled within a range.  Previous data in the range will be overwritten.
        """
        command = f"DATA,{ir},{lstrt},{lstop},{linc},{name},{kcplx}"
        return self.run(command, **kwargs)

    def edread(self, nstart="", label="", num="", step1="", step2="", **kwargs):
        """Reads explicit dynamics output into variables for time-history

        APDL Command: EDREAD
        postprocessing.

        Parameters
        ----------
        nstart
            Starting reference number assigned to the first variable. Allowed
            range is 2 (the default) to NV [NUMVAR].  (NV defaults to 30 for an
            explicit dynamics analysis.)

        label
            Label identifying the output file to be read. No default.

            GLSTAT - Read data from the GLSTAT file.

            MATSUM - Read data from the MATSUM file.

            SPCFORC - Read data from the SPCFORC file.

            RCFORC - Read data from the RCFORC file.

            SLEOUT - Read data from the SLEOUT file.

            NODOUT - Read data from the NODOUT file.

            RBDOUT - Read data from the RBDOUT file.

        num
            Number identifying the data set to be read in (defaults to 1). If
            Label = GLSTAT, NUM is ignored. If Label = MATSUM or RBDOUT, NUM is
            the PART number [EDPART] for which output is desired. If Label =
            SPCFORC or NODOUT, NUM is the node number for which output is
            desired. If Label  = SLEOUT or RCFORC, NUM is the number of the
            contact entity for which output is desired.

        step1, step2
            Load step range of data to be read in. If STEP1 and STEP2 are
            blank, all load steps are read in.

        Notes
        -----
        EDREAD reads data from the specified ascii output file so that it may
        be used during postprocessing. After EDREAD, you must issue the STORE
        command to store the data in time history variables. Once stored, the
        variables can be viewed as plots of output item versus time.

        The number of variables stored depends on the file specified. The
        following table shows the items in each file and the order in which
        they are stored.  If data items were previously stored in variables
        NSTART to NSTART+15, they will be overwritten. If more variables are
        needed, change NV on the NUMVAR command. (Note that hourglass energy
        will not be available if it was not specified for the model
        [EDENERGY,1].)

        The following items under MATSUM are listed in the MATSUM ASCII file
        (in the Mat no. field) for each part number at time intervals specified
        by the EDHTIME command. Use EDREAD,,MATSUM,NUM to specify the part
        number that corresponds to the mat number in the MATSUM file.

        Resultant contact forces and sliding interface energies are available
        from the RCFORC and SLEOUT files, respectively. The RCFORC file is
        written for surface based contact types that include target and contact
        (master and slave) definitions.  You should ensure that this file
        contains valid force results before issuing EDREAD,,RCFORC. Only the
        resultant contact forces on the master surface are available for time-
        history postprocessing.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"EDREAD,{nstart},{label},{num},{step1},{step2}"
        return self.run(command, **kwargs)

    def enersol(self, nvar="", item="", name="", **kwargs):
        """Specifies the total energies to be stored.

        APDL Command: ENERSOL

        Parameters
        ----------
        nvar
            Arbitrary reference number assigned to this variable (2 to NV).
        """
        command = f"ENERSOL,{nvar},{item},{name}"
        return self.run(command, **kwargs)

    def esol(
        self,
        nvar: MapdlInt = "",
        elem: MapdlInt = "",
        node: MapdlInt = "",
        item: str = "",
        comp: str = "",
        name: str = "",
        **kwargs,
    ) -> Optional[str]:
        """Specify element data to be stored from the results file.

        /POST26 APDL Command: ESOL

        Parameters
        ----------
        nvar
            Arbitrary reference number assigned to this variable (2 to
            NV [NUMVAR]). Overwrites any existing results for this
            variable.

        elem
            Element for which data are to be stored.

        node
            Node number on this element for which data are to be
            stored. If blank, store the average element value (except
            for FMAG values, which are summed instead of averaged).

        item
            Label identifying the item. General item labels are shown
            in Table 134: ESOL - General Item and Component Labels
            below. Some items also require a component label.

        comp
            Component of the item (if required). General component
            labels are shown in Table 134: ESOL - General Item and
            Component Labels below.  If Comp is a sequence number (n),
            the NODE field will be ignored.

        name
            Thirty-two character name for identifying the item on the
            printout and displays.  Defaults to a label formed by
            concatenating the first four characters of the Item and
            Comp labels.

        Notes
        -----
        See Table: 134:: ESOL - General Item and Component Labels for
        a list of valid item and component labels for element (except
        line element) results.

        The ESOL command defines element results data to be stored
        from a results file (FILE). Not all items are valid for all
        elements. To see the available items for a given element,
        refer to the input and output summary tables in the
        documentation for that element.

        Two methods of data access are available via the ESOL
        command. You can access some simply by using a generic label
        (component name method), while others require a label and
        number (sequence number method).

        Use the component name method to access general element data
        (that is, element data generally available to most element
        types or groups of element types).

        The sequence number method is required for data that is not
        averaged (such as pressures at nodes and temperatures at
        integration points), or data that is not easily described in a
        generic fashion (such as all derived data for structural line
        elements and contact elements, all derived data for thermal
        line elements, and layer data for layered elements).

        Element results are in the element coordinate system, except
        for layered elements where results are in the layer coordinate
        system.  Element forces and moments are in the nodal
        coordinate system. Results are obtainable for an element at a
        specified node. Further location specifications can be made
        for some elements via the SHELL, LAYERP26, and FORCE commands.

        For more information on the meaning of contact status and its
        possible values, see Reviewing Results in POST1 in the Contact
        Technology Guide.

        Examples
        --------
        Switch to the time-history postprocessor

        >>> mapdl.post26()

        Store the stress in the X direction for element 1 at node 1

        >>> nvar = 2
        >>> mapdl.esol(nvar, 1, 1, 'S', 'X')

        Move the value to an array and access it via mapdl.parameters

        >>> mapdl.dim('ARR', 'ARRAY', 1)
        >>> mapdl.vget('ARR', nvar)
        >>> mapdl.parameters['ARR']
        array(-1991.40234375)

        """
        command = f"ESOL,{nvar},{elem},{node},{item},{comp},{name}"
        return self.run(command, **kwargs)

    def gapf(self, nvar="", num="", name="", **kwargs):
        """Defines the gap force data to be stored in a variable.

        APDL Command: GAPF

        Parameters
        ----------
        nvar
            Arbitrary reference number assigned to this variable (2 to NV
            [NUMVAR]).  Overwrites any existing results for this variable.

        num
            Number identifying gap number for which the gap force is to be
            stored.  Issue the GPLIST command to display gap numbers.

        name
            Thirty-two character name for identifying the item on the printout
            and displays (defaults to the name GAPF).

        Notes
        -----
        Defines the gap force data to be stored in a variable. Applicable only
        to the expansion pass of the mode-superposition linear transient
        dynamic (ANTYPE,TRANS) analysis. The data is usually on Fname.RDSP.
        """
        command = f"GAPF,{nvar},{num},{name}"
        return self.run(command, **kwargs)

    def gssol(self, nvar="", item="", comp="", name="", **kwargs):
        """Specifies which results to store from the results file when using

        APDL Command: GSSOL
        generalized plane strain.

        Parameters
        ----------
        nvar
            Arbitrary reference number or name assigned to this variable.
            Variable numbers can be 2 to NV (NUMVAR) while the name can be an
            eight byte character string. Overwrites any existing results for
            this variable.

        item
            Label identifying item to be stored.

            LENGTH - Change of fiber length at the ending point.

            ROT - Rotation of the ending plane during deformation.

            F - Reaction force at the ending point in the fiber direction.

            M - Reaction moment applied on the ending plane.

        comp
            Component of the item, if Item = ROT or M.

            X - The rotation angle or reaction moment of the ending plane about X.

            Y - The rotation angle or reaction moment of the ending plane about Y.

        name
            Thirty-two character name identifying the item on the printout and
            display. Defaults to the label formed by concatenating the first
            four characters of the Item and Comp labels.

        Notes
        -----
        This command stores the results (new position of the ending plane after
        deformation) for generalized plane strain. All outputs are in the
        global Cartesian coordinate system. For more information about the
        generalized plane strain feature, see Generalized Plane Strain Option
        of Current-Technology Solid Elements in the Element Reference.
        """
        command = f"GSSOL,{nvar},{item},{comp},{name}"
        return self.run(command, **kwargs)

    def jsol(self, nvar="", elem="", item="", comp="", name="", **kwargs):
        """Specifies result items to be stored for the joint element.

        APDL Command: JSOL

        Parameters
        ----------
        nvar
            Arbitrary reference number or name assigned to this variable.
            Variable numbers can be 2 to NV (NUMVAR) while the name can be an
            eight-byte character string. Overwrites any existing results for
            this variable.

        elem
            Element number for which to store results.

        item
            Label identifying the item.  Valid item labels are shown in
            Table 202: JSOL - Valid Item and Component Labels below.

        comp
            Component of the Item (if required).  Valid component labels are
            shown in Table 202: JSOL - Valid Item and Component Labels below.

        name
            Thirty-two character name identifying the item on printouts and
            displays.  Defaults to a label formed by concatenating the first
            four characters of the Item and Comp labels.

        Notes
        -----
        This command is valid for the MPC184 joint elements. The values stored
        are for the free or unconstrained degrees of freedom of a joint
        element. Relative reaction forces and moments are available only if
        stiffness, damping, or friction is associated with the joint element.

        Table: 202:: : JSOL - Valid Item and Component Labels


        """
        command = f"JSOL,{nvar},{elem},{item},{comp},{name}"
        return self.run(command, **kwargs)

    def nsol(self, nvar="", node="", item="", comp="", name="", sector="", **kwargs):
        """Specifies nodal data to be stored from the results file.

        APDL Command: NSOL

        Parameters
        ----------
        nvar
            Arbitrary reference number or name assigned to this variable.
            Variable numbers can be 2 to NV (NUMVAR) while the name can be an
            eight byte character string. Overwrites any existing results for
            this variable.

        node
            Node for which data are to be stored.

        item
            Label identifying the item.  Valid item labels are shown in the
            table below.  Some items also require a component label.

        comp
            Component of the item (if required).  Valid component labels are
            shown in the table below.

        name
            Thirty-two character name identifying the item on printouts and
            displays.  Defaults to a label formed by concatenating the first
            four characters of the Item and Comp labels.

        sector
            For a full harmonic cyclic symmetry solution, the sector number for
            which the results from NODE are to be stored.

        Notes
        -----
        Stores nodal degree of freedom and solution results in a variable. For
        more information, see Data Interpreted in the Nodal Coordinate System
        in the Modeling and Meshing Guide.

        For SECTOR>1, the result is in the nodal coordinate system of the base
        sector, and it is rotated to the expanded sector’s location. Refer to
        Using the /CYCEXPAND Command in the Cyclic Symmetry Analysis Guide for
        more information.

        For SHELL131 and SHELL132 elements with KEYOPT(3) = 0 or 1, use the
        labels TBOT, TE2, TE3, . . ., TTOP instead of TEMP.
        """
        command = f"NSOL,{nvar},{node},{item},{comp},{name},{sector}"
        return self.run(command, **kwargs)

    def nstore(self, tinc="", **kwargs):
        """Defines which time points are to be stored.

        APDL Command: NSTORE

        Parameters
        ----------
        tinc
            Store data associated with every TINC time (or frequency) point(s),
            within the previously defined range of TMIN to TMAX [TIMERANGE].
            (Defaults to 1)

        Notes
        -----
        Defines which time (or frequency) points within the range are to be
        stored.
        """
        command = f"NSTORE,{tinc}"
        return self.run(command, **kwargs)

    def numvar(self, nv="", **kwargs):
        """Specifies the number of variables allowed in POST26.

        APDL Command: NUMVAR

        Parameters
        ----------
        nv
            Allow storage for NV variables.  200 maximum are allowed.  Defaults
            to 10 (except for an explicit dynamics analysis, which defaults to
            30).  TIME (variable 1) should also be included in this number.

        Notes
        -----
        Specifies the number of variables allowed for data read from the
        results file and for data resulting from an operation (if any).  For
        efficiency, NV should not be larger than necessary.  NV cannot be
        changed after data storage begins.
        """
        command = f"NUMVAR,{nv}"
        return self.run(command, **kwargs)

    def reset(self, **kwargs):
        """Resets all POST1 or POST26 specifications to initial defaults.

        APDL Command: RESET

        Notes
        -----
        Has the same effect as entering the processor the first time within the
        run.  In POST1, resets all specifications to initial defaults, erases
        all element table items, path table data, fatigue table data, and load
        case pointers.  In POST26, resets all specifications to initial
        defaults, erases all variables defined, and zeroes the data storage
        space.
        """
        command = f"RESET,"
        return self.run(command, **kwargs)

    def rforce(self, nvar="", node="", item="", comp="", name="", **kwargs):
        """Specifies the total reaction force data to be stored.

        APDL Command: RFORCE

        Parameters
        ----------
        nvar
            Arbitrary reference number assigned to this variable (2 to NV
            [NUMVAR]). Overwrites any existing results for this variable.

        node
            Node for which data are to be stored.  If NODE = P, graphical
            picking is enabled (valid only in the GUI).

        item
            Label identifying the item.  Valid item labels are shown in the
            table below.  Some items also require a component label.

        comp
            Component of the item (if required).  Valid component labels are
            shown in the table below.

        name
            Thirty-two character name identifying the item on printouts and
            displays. Defaults to an eight character label formed by
            concatenating the first four characters of the Item and Comp
            labels.

        Notes
        -----
        Defines the total reaction force data (static, damping, and inertial
        components) to be stored from single pass (ANTYPE,STATIC or TRANS)
        solutions or from the expansion pass of mode-superposition
        (ANTYPE,HARMIC or TRANS) solutions.

        Table: 228:: : RFORCE - Valid Item and Component Labels

        For SHELL131 and SHELL132 elements with KEYOPT(3) = 0 or 1, use the
        labels HBOT, HE2, HE3, . . ., HTOP instead of HEAT.
        """
        command = f"RFORCE,{nvar},{node},{item},{comp},{name}"
        return self.run(command, **kwargs)

    def rgb(
        self,
        kywrd="",
        pred="",
        pgrn="",
        pblu="",
        n1="",
        n2="",
        ninc="",
        ncntr="",
        **kwargs,
    ):
        """Specifies the RGB color values for indices and contours.

        APDL Command: /RGB

        Parameters
        ----------
        kywrd
            Determines how RGB modifications will be applied.

            INDEX - Specifies that subsequent color values apply to ANSYS color indices (0-15).

            CNTR - Specifies that subsequent color values apply to contours (1-128).  Applies to
                   C-option devices only (i.e. X11C or Win32C).

        pred
            Intensity of the color red, expressed as a percentage.

        pgrn
            Intensity of the color green, expressed as a percentage.

        pblu
            Intensity of the color blue, expressed as a percentage.

        n1
            First index (0-15), or contour (1-128) to which the designated RGB
            values apply.

        n2
            Final index (0-15), or contour (1-128) to which the designated RGB
            values apply.

        ninc
            The step increment between the values N1 and N2 determining which
            contours or indices will be controlled by the specified RGB values.

        ncntr
            The new maximum number of contours (1-128).

        Notes
        -----
        Issuing the /CMAP command (with no filename) will restore the default
        color settings.
        """
        command = f"/RGB,{kywrd},{pred},{pgrn},{pblu},{n1},{n2},{ninc},{ncntr}"
        return self.run(command, **kwargs)

    def solu(self, nvar="", item="", comp="", name="", **kwargs):
        """Specifies solution summary data per substep to be stored.

        APDL Command: SOLU

        Parameters
        ----------
        nvar
            Arbitrary reference number assigned to this variable (2 to NV
            [NUMVAR]).

        item
            Label identifying the item.  Valid item labels are shown in the
            table below.  Some items may also require a component label.

        comp
            Component of the item (if required).  Valid component labels are
            shown in the table below.  None are currently required.

        name
            Thirty-two character name identifying the item on printouts and
            displays.  Defaults to an eight character label formed by
            concatenating the first four characters of the Item and Comp
            labels.

        Notes
        -----
        See also the PRITER command of POST1 to display some of these items
        directly.  Valid for a static or full transient analysis. All other
        analyses have zeros for the data. Valid item and component labels for
        solution summary values are:
        """
        command = f"SOLU,{nvar},{item},{comp},{name}"
        return self.run(command, **kwargs)

    def store(self, lab="", npts="", **kwargs):
        """Stores data in the database for the defined variables.

        APDL Command: STORE

        Parameters
        ----------
        lab
            Valid labels:

            MERGE - Merge data from results file for the time points in memory with the existing
                    data using current specifications (default).

            NEW - Store a new set of data, replacing any previously stored data with current
                  result file specifications and deleting any previously-
                  calculated (OPER) variables. Variables defined using the
                  ANSOL command are also deleted.

            APPEN - Append data from results file to the existing data.

            ALLOC - Allocate (and zero) space for NPTS data points.

            PSD - Create a new set of frequency points for PSD calculations (replacing any
                  previously stored data and erasing any previously calculated
                  data).

        npts
            The number of time points (or frequency points) for storage (used
            only with Lab = ALLOC or PSD).  The value may be input when using
            POST26 with data supplied from other than a results file.  This
            value is automatically determined from the results file data with
            the NEW, APPEN, and MERGE options.  For the PSD option, NPTS
            determines the resolution of the frequency vector (valid numbers
            are between 1 and 10, defaults to 5).

        Notes
        -----
        This command stores data from the results file in the database for the
        defined variables [NSOL, ESOL, SOLU, JSOL] per specification [FORCE,
        LAYERP26, SHELL].  See the Basic Analysis Guide for more information.

        The STORE,PSD command will create a new frequency vector (variable 1)
        for response PSD calculations [RPSD].  This command should first be
        issued before defining variables [NSOL, ESOL, RFORCE] for which
        response PSD's are to be calculated.
        """
        command = f"STORE,{lab},{npts}"
        return self.run(command, **kwargs)

    def timerange(self, tmin="", tmax="", **kwargs):
        """Specifies the time range for which data are to be stored.

        APDL Command: TIMERANGE

        Parameters
        ----------
        tmin
            Minimum time (defaults to first time (or frequency) point on the
            file).

        tmax
            Maximum time (defaults to last time (or frequency) point on the
            file).

        Notes
        -----
        Defines the time (or frequency) range for which data are to be read
        from the file and stored in memory.  Use the NSTORE command to define
        the time increment.

        Use PRTIME or PLTIME to specify the time (frequency) range for cyclic
        mode-superposition harmonic analyses.
        """
        command = f"TIMERANGE,{tmin},{tmax}"
        return self.run(command, **kwargs)

    def vardel(self, nvar="", **kwargs):
        """Deletes a variable (GUI).

        APDL Command: VARDEL

        Parameters
        ----------
        nvar
            The reference number of the variable to be deleted.  NVAR is as
            defined by NSOL, ESOL, etc.

        Notes
        -----
        Deletes a POST26 solution results variable.  This is a command
        generated by the Graphical User Interface (GUI).  It will appear in the
        log file (Jobname.LOG) if a POST26 variable is deleted from the
        "Defined Time-History Variables" dialog box.  This command is not
        intended to be typed in directly in an ANSYS session (although it can
        be included in an input file for batch input or for use with the /INPUT
        command).
        """
        command = f"VARDEL,{nvar}"
        return self.run(command, **kwargs)

    def varnam(self, ir="", name="", **kwargs):
        """Names (or renames) a variable.

        APDL Command: VARNAM

        Parameters
        ----------
        ir
            Reference number of the variable (2 to NV [NUMVAR]).

        name
            Thirty-two character name for identifying variable on printouts and
            displays.  Embedded blanks are compressed for output.
        """
        command = f"VARNAM,{ir},{name}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class LoadStepOptions:
    def autots(self, key="", **kwargs):
        """Specifies whether to use automatic time stepping or load stepping.

        APDL Command: AUTOTS

        Parameters
        ----------
        key
            Automatic time stepping key:

            OFF
                Do not use automatic time stepping.

            ON
                Use automatic time stepping (default).

            AUTO
                The program determines whether to use automatic time stepping (used by
                Workbench).

        Notes
        -----
        Specifies whether to use automatic time stepping (or load stepping)
        over this load step. If Key = ON, both time step prediction and time
        step bisection will be used.

        You cannot use automatic time stepping ``AUTOTS``, line search ``LNSRCH``,
        or the DOF solution predictor [PRED] with the arc-length method
        ``ARCLEN, ARCTRM``. If you activate the arc-length method after you set
        ``AUTOTS, LNSRCH``, or ``PRED``, a warning message appears. If you choose to
        proceed with the arc-length method, the program disables your automatic
        time stepping, line search, and DOF predictor settings, and the time
        step size is controlled by the arc-length method internally.

        This command is also valid in PREP7.
        """
        command = f"AUTOTS,{key}"
        return self.run(command, **kwargs)

    def campbell(self, action="", **kwargs):
        """Prepares the result file for a subsequent Campbell diagram of a

        APDL Command: CAMPBELL
        prestressed structure.

        Parameters
        ----------
        action
            Campbell action:

            NONE - Do not prepare the result file. This option is the default behavior.

            RSTP - Prepare the result file (Jobname.RSTP) for a subsequent Campbell diagram of a
                   prestressed structure.

        Notes
        -----
        For an analysis involving a prestressed structure, the ``CAMPBELL`` command
        specifies whether or not to prepare the result file to support  a
        Campbell diagram analysis (``PRCAMP`` or ``PLCAMP``).

        To prestress a structure, the program performs a static solution before
        the linear perturbation modal solution.

        The CAMPBELL command requires that modal and static analyses be
        performed alternately. It works only when the number of static analyses
        is the same as the number of modal analyses. Any number of analyses can
        be performed, but the same number of each (static and modal) is
        expected. The modal solutions are appended in the results file
        (Jobname.RSTP).

        For an example of PLCAMP command usage, see Example Campbell Diagram
        Analysis in the Advanced Analysis Guide.
        """
        command = f"CAMPBELL,{action}"
        return self.run(command, **kwargs)

    def cecmod(self, neqn="", const="", **kwargs):
        """Modifies the constant term of a constraint equation during solution.

        APDL Command: CECMOD

        Parameters
        ----------
        neqn
            Reference number of constraint equation.

        const
            New value of the constant term of equation.

        Notes
        -----
        Other terms of the constraint equation cannot be changed during the
        solution phase, but must be defined or changed within PREP7 prior to
        the solution.  See the CE command for details.

        This command is also valid in PREP7.
        """
        command = f"CECMOD,{neqn},{const}"
        return self.run(command, **kwargs)

    def deltim(self, dtime="", dtmin="", dtmax="", carry="", **kwargs):
        """Specifies the time step sizes to be used for the current load step.

        APDL Command: DELTIM

        Parameters
        ----------
        dtime
            Time step size for this step. If automatic time stepping is used
            (AUTOTS), DTIME is the starting time substep.

        dtmin
            Minimum time step (if automatic time stepping is used). The program
            automatically determines the default based on the physics of the
            model.

        dtmax
            Maximum time step (if automatic time stepping is used). The program
            automatically determines the default based on the physics of the
            model.

        carry
            Time step carry over key:

            OFF - Use DTIME as time step at start of each load step.

            ON - Use final time step from previous load step as the starting time step (if
                 automatic time stepping is used).

        Notes
        -----
        See NSUBST for an alternative input.

        Use consistent values for DTIME and TIME (TIME). For example, using 0.9
        for DTIME and 1.0 for TIME results in one time step because 1.0 (TIME)
        is divisible by .9 (DTIME) at most once. If you intend to load in 10
        increments over a time span of 1.0, use 0.1 for DTIME and 1.0 for TIME.

        The program calculates the initial incremental time so that (EndingTime
        - StartingTime)/DTIME is an integer, which may affect the initial
        incremental time that you specify. For example, if the starting time is
        0, the ending time is 1, and the initial incremental time is 0.4, the
        program rounds to the nearest integer and adjusts the time to 0.33333.

        For solution efficiency, specify values for all fields of this command.

        This command is also valid in PREP7.
        """
        command = f"DELTIM,{dtime},{dtmin},{dtmax},{carry}"
        return self.run(command, **kwargs)

    def expsol(self, lstep="", sbstep="", timfrq="", elcalc="", **kwargs):
        """Specifies the solution to be expanded for mode-superposition analyses

        APDL Command: EXPSOL
        or substructure analyses.

        Parameters
        ----------
        lstep, sbstep
            Expand the solution identified as load step LSTEP and substep
            SBSTEP.

        timfrq
            As an alternative to LSTEP and SBSTEP, expand the solution at, or
            nearest to, the time value TIMFRQ (for ANTYPE,TRANS or
            ANTYPE,SUBSTR) or frequency value TIMFRQ (for ANTYPE,HARMIC).
            LSTEP and SBSTEP should be blank.

        elcalc
            Element calculation key:

            YES - Calculate element results, nodal loads, and reaction loads.

            NO - Do not calculate these items.

        Notes
        -----
        Specifies the solution to be expanded from analyses that use the mode-
        superposition method (ANTYPE,HARMIC or TRANS) or substructuring
        (ANTYPE,SUBSTR). Use the NUMEXP command to expand a group of solutions.

        The resulting results file will maintain the same load step, substep,
        and time (or frequency) values as the requested solution to be
        expanded.

        This command is also valid in PREP7.
        """
        command = f"EXPSOL,{lstep},{sbstep},{timfrq},{elcalc}"
        return self.run(command, **kwargs)

    def kbc(self, key="", **kwargs):
        """Specifies ramped or stepped loading within a load step.

        APDL Command: KBC

        Parameters
        ----------
        key
            Ramping key:

            0 - Loads are linearly interpolated (ramped) for each substep from the values of
                the previous load step to the values of this load step. This is
                the default value.

            1 - Loads are step changed (stepped) at the first substep of this load step to the
                values of this load step (i.e., the same values are used for
                all substeps).  Useful for rate-dependent behavior (e.g.,
                creep, viscoplasticity, etc.) or transient load steps only.

        Notes
        -----
        Specifies whether loads applied to intermediate substeps within the
        load step are to be stepped or ramped. Used only if DTIME on the DELTIM
        command is less than the time span or, conversely, if NSBSTP on the
        NSUBST command is greater than one. Flags (FSI, MXWF, MVDI, etc.) are
        always stepped.

        Changing the ramping KEY (i.e., switching between ramped and stepped
        boundary conditions) between load steps is not recommended.

        For ramped loading (KBC,0), when a load is applied for the first time,
        it is interpolated from zero to the value of the current load step, and
        not from the initial condition or value of the degree of freedom from
        the previous load step.

        Spatially varying tabular loads or boundary conditions do not support
        direct ramping or stepping options and, instead, apply their full
        values according to the supplied tabular functions regardless of the
        KBC setting.

        For a static or harmonic cyclic symmetry analysis, any load that varies
        by sector (CYCOPT,LDSECT) is tabular and is applied as a step change,
        regardless of the KBC setting; however, any non-tabular loads in the
        same analysis are ramped or stepped according to the KBC setting.

        Irrespective of the KBC setting, loads are usually step-removed. See
        Stepping or Ramping Loads in the Basic Analysis Guide for more
        information.

        It is sometimes difficult to obtain successful convergence with stepped
        loading in a nonlinear transient problem. If divergence is encountered,
        determine if stepped loading was used by default, then determine if it
        is appropriate for the analysis.

        This command is also valid in PREP7.
        """
        command = f"KBC,{key}"
        return self.run(command, **kwargs)

    def kuse(self, key="", **kwargs):
        """Specifies whether or not to reuse the factorized matrix.

        APDL Command: KUSE

        Parameters
        ----------
        key
            Reuse key:

            - 0 : Program decides whether or not to reuse the previous
              factorized stiffness matrix.
            - 1 : Force the previous factorized stiffness matrix to be
              reused.  Used mainly in a restart.  Forcing reuse of the
              matrix is a nonstandard use of the program, and
              should be done with caution.  For instance, using
              this option and changing the number of elements, or
              the number or type of degrees of freedom, may cause
              an abort.
            - -1 : All element matrices are reformed and are used to
              reform a new factorized stiffness matrix.

        Notes
        -----
        Overrides the program logic to determine whether or not to reuse the
        previous factorized stiffness matrix for each substep of this load
        step.  Applies only to static or full transient analyses and to full
        harmonic analyses if the frequency is not changed for continuing
        loadsteps. For full harmonic analyses, only KEY = 1 or KEY = 0 is
        valid.

        This command is also valid in PREP7.
        """
        command = f"KUSE,{key}"
        return self.run(command, **kwargs)

    def magopt(self, value="", **kwargs):
        """Specifies options for a 3-D magnetostatic field analysis.

        APDL Command: MAGOPT

        Parameters
        ----------
        value
            Option key:

            0 - Calculate a complete H field solution in the entire domain using a single
                (reduced) potential.

            Caution:When used in problems with both current sources and iron regions, errors may result due to numerical cancellation. - 1

            Calculate and store a preliminary H field in "iron" regions (μr ≠ 1).  Requires flux-parallel boundary conditions to be specified on exterior iron boundaries.  Used in conjunction with subsequent solutions with VALUE = 2 followed by VALUE = 3.  Applicable to multiply-connected iron domain problems. - 2

            Calculate and store a preliminary H field in "air" regions (μr = 1).  The air-iron interface is appropriately treated internally by the program.  Used in conjunction with a subsequent solution with VALUE = 3.  Applicable to singly-connected iron domain problems (with subsequent solution with VALUE = 3) or to multiply-connected iron domain problems (when preceded by a solution with VALUE = 1 and followed by a solution with VALUE = 3). - 3

        Notes
        -----
        Specifies the solution sequence options for a 3-D magnetostatic field
        analysis using a scalar potential (MAG).  The solution sequence is
        determined by the nature of the problem.

        You cannot use constraint equations with Value = 1.

        This command is also valid in PREP7.

        Distributed ANSYS Restriction: The MAGOPT,3 option is not supported in
        Distributed ANSYS when the following contact elements are present in
        the model: CONTA173, CONTA174, CONTA175, CONTA176, or CONTA177.
        """
        command = f"MAGOPT,{value}"
        return self.run(command, **kwargs)

    def magsolv(
        self,
        opt="",
        nramp="",
        cnvcsg="",
        cnvflux="",
        neqit="",
        biot="",
        cnvtol="",
        **kwargs,
    ):
        """Specifies magnetic solution options and initiates the solution.

        APDL Command: MAGSOLV

        Parameters
        ----------
        opt
            Static magnetic solution option:

            0 - Vector potential (MVP) or edge formulation  (default).

            1 - Combined vector potential and reduced scalar potential (MVP-RSP).

            2 - Reduced scalar potential (RSP).

            3 - Difference scalar potential (DSP).

            4 - General scalar potential (GSP).

        nramp
            Number of ramped substeps for the first load step of a nonlinear
            MVP or MVP-RSP solution.  Defaults to 3.  If NRAMP = -1, ignore the
            ramped load step entirely.NRAMP is ignored for linear
            magnetostatics.

        cnvcsg
            Tolerance value on the program-calculated reference value for the
            magnetic current-segment convergence.  Used for the MVP, the MVP-
            RSP, and the edge formulation solution options (OPT = 0 and 1).
            Defaults to 0.001.

        cnvflux
            Tolerance value on the program-calculated reference value for the
            magnetic flux convergence.  Used for all scalar potential solution
            options (OPT = 2, 3, 4).  Defaults to 0.001.

        neqit
            Maximum number of equilibrium iterations per load step.  Defaults
            to 25.

        biot
            Option to force execution of a Biot-Savart integral solution
            [BIOT,NEW] for the scalar potential options.  Required if multiple
            load steps are being performed with different current source
            primitives (SOURC36 elements).

            0 - Do not force execution of Biot-Savart calculation (default); Biot-Savart is
                automatically calculated only for the first solution.

            1 - Force execution of Biot-Savart calculation.

        cnvtol
            Sets the convergence tolerance for AMPS reaction. Defaults to 1e-3.

        Notes
        -----
        MAGSOLV invokes an ANSYS macro which specifies magnetic solution
        options and initiates the solution.  The macro is applicable to any
        ANSYS magnetostatic analysis using the magnetic vector potential (MVP),
        reduced scalar potential (RSP), difference scalar potential (DSP),
        general scalar potential (GSP), or combined MVP-RSP formulation
        options.  Results are only stored for the final converged solution.
        (In POST1, issue ``*SET,LIST`` to identify the load step of solution
        results.)  The macro internally determines if a nonlinear analysis is
        required based on magnetic material properties.

        If you use the BIOT option and issue SAVE after solution or
        postprocessing, the Biot-Savart calculations are saved to the database,
        but will be overwritten upon normal exit from the program.  To save
        this data after issuing SAVE, use the /EXIT,NOSAVE command.  You can
        also issue the /EXIT,SOLU command to exit ANSYS and save all solution
        data, including the Biot-Savart calculations, in the database.
        Otherwise, when you issue RESUME, the Biot-Savart calculation will be
        lost (resulting in a zero solution).

        The MVP, MVP-RSP, and edge formulation options perform a two-load-step
        solution sequence.  The first load step ramps the applied loads over a
        prescribed number of substeps (NRAMP), and the second load step
        calculates the converged solution.  For linear problems, only a single
        load step solution is performed.  The ramped load step can be bypassed
        by setting NRAMP to -1.

        The RSP option solves in a single load step using the adaptive descent
        procedure.  The DSP option uses two load steps, and the GSP solution
        uses three load steps.

        The following analysis options and nonlinear options are controlled by
        this macro:  KBC, NEQIT, NSUBST, CNVTOL, NROPT, MAGOPT, and OUTRES.

        You cannot use constraint equations with OPT = 4.
        """
        command = f"MAGSOLV,{opt},{nramp},{cnvcsg},{cnvflux},{neqit},{biot},{cnvtol}"
        return self.run(command, **kwargs)

    def mode(self, mode="", isym="", **kwargs):
        """Specifies the harmonic loading term for this load step.

        APDL Command: MODE

        Parameters
        ----------
        mode
            Number of harmonic waves around circumference for this harmonic
            loading term (defaults to 0).

        isym
            Symmetry condition for this harmonic loading term (not used when
            MODE = 0):

            1 - Symmetric (UX, UY, ROTZ, TEMP use cosine terms; UZ uses sine term) (default).

            -1 - Antisymmetric (UX, UY, ROTZ, TEMP use sine terms; UZ uses cosine term).

        Notes
        -----
        Used with axisymmetric elements having nonaxisymmetric loading
        capability (for example, PLANE25, SHELL61, etc.).  For analysis types
        ANTYPE,MODAL, HARMIC, TRANS, and SUBSTR, the term must be defined in
        the first load step and may not be changed in succeeding load steps.

        This command is also valid in PREP7.
        """
        command = f"MODE,{mode},{isym}"
        return self.run(command, **kwargs)

    def nsubst(self, nsbstp="", nsbmx="", nsbmn="", carry="", **kwargs):
        """Specifies the number of substeps to be taken this load step.

        APDL Command: NSUBST

        Parameters
        ----------
        nsbstp
            Number of substeps to be used for this load step (i.e., the time
            step size or frequency increment). If automatic time stepping is
            used (``AUTOTS``), ``NSBSTP`` defines the size of the first substep.

        nsbmx
            Maximum number of substeps to be taken (i.e., the minimum time step
            size) if automatic time stepping is used. The program automatically
            determines the default based on the physics of the model.

        nsbmn
            Minimum number of substeps to be taken (i.e., the maximum time step
            size) if automatic time stepping is used. The program automatically
            determines the default based on the physics of the model.

        carry
            Time step carryover key (program-determined default depending on
            the problem physics):

            OFF
                Use NSBSTP to define time step at start of each load step.

            ON
                Use final time step from previous load step as the starting time step (if
                automatic time stepping is used).

        Notes
        -----
        See ``DELTIM`` for an alternative input.  It is recommended that all fields
        of this command be specified for solution efficiency and robustness.

        When the arc-length method is active (``ARCLEN`` command), the ``NSBMX`` and
        ``NSBMN`` arguments are ignored.

        This command is also valid in PREP7.
        """
        command = f"NSUBST,{nsbstp},{nsbmx},{nsbmn},{carry}"
        return self.run(command, **kwargs)

    def numexp(self, num="", begrng="", endrng="", elcalc="", **kwargs):
        """Specifies solutions to be expanded from mode-superposition analyses or

        APDL Command: NUMEXP
        substructure analyses.

        Parameters
        ----------
        num
            The number of solutions to expand. This value is required.

            Num
                Number of solutions to expand.

            ALL
                Expand all substeps between ``BEGRNG`` and ``ENDRNG`` (provided that ENDRNG > 0). If
                ``BEGRNG`` and ``ENDRNG`` have no specified values, this option
                expands all substeps of all load steps.

        begrng, endrng
            Beginning and ending time (or frequency) range for expanded
            solutions. The default is 0 for both values.

        elcalc
            The element-calculation key:

            YES
                Calculate element results, nodal loads, and reaction loads. This value is the
                default.

            NO
                Do not calculate these items.

        Notes
        -----
        Specifies a range of solutions to be expanded from analyses that use
        mode-superposition methods (``ANTYPE,HARMIC`` or ``TRANS``) or substructuring
        (``ANTYPE,SUBSTR``).

        For ANTYPE,TRANS, NUM, evenly spaced solutions are expanded between
        time BEGRNG and time ``ENDRNG``.

        For ANTYPE,HARMIC, NUM, evenly spaced solutions are expanded between
        frequency BEGRNG and frequency ``ENDRNG``.

        The first expansion in all cases is done at the first point beyond
        BEGRNG (that is, at ``BEGRNG + (ENDRNG - BEGRNG) / NUM``)).

        The resulting results file will maintain the same load step, substep,
        and time (or frequency) values as the use pass.

        For a single expansion of a solution, or for multiple expansions when
        the solutions are not evenly spaced (such as in a mode-superposition
        harmonic analysis with the cluster option), ANSYS, Inc. recommends
        issuing one or more ``EXPSOL`` commands.

        The NUMEXP command is invalid in these cases:

        In a substructing analysis (``ANTYPE,SUBST``) when a factorized matrix file
        (the ``.LN22`` file generated by the sparse solver) does not exist, causing
        ANSYS to employ the full-resolve method.

        If the full-resolve option is selected using the ``SEOPT`` command.

        In both situations, use the ``EXPSOL`` command to perform a single
        expansion for each solution desired.

        This command is also valid in PREP7.
        """
        command = f"NUMEXP,{num},{begrng},{endrng},{elcalc}"
        return self.run(command, **kwargs)

    def time(self, time="", **kwargs):
        """Sets the time for a load step.

        APDL Command: TIME

        Parameters
        ----------
        time
            Time at the end of the load step.

        Notes
        -----
        Associates the boundary conditions at the end of the load step with a
        particular TIME value.

        TIME must be a positive, nonzero, monotonically increasing quantity
        that "tracks" the input history.  Units of time should be consistent
        with those used elsewhere (for properties, creep equations, etc.).

        Typically, for the first load step TIME defaults to 1. However, for the
        first load step of a mode-superposition transient analysis
        (ANTYPE,TRANS and TRNOPT,MSUP), the TIME command is ignored and a
        static solution is performed at TIME = 0.

        For a full transient analyses, the command's default behavior does not
        apply. You must specify a time for each load step and it must be
        greater than the time at the end of the prior load step.

        TIME does not apply to modal (ANTYPE,MODAL), harmonic (ANTYPE,HARMIC),
        or substructure (ANTYPE,SUBSTR) analyses.

        This command is also valid in PREP7.
        """
        command = f"TIME,{time}"
        return self.run(command, **kwargs)

    def tref(self, tref="", **kwargs):
        """Defines the reference temperature for the thermal strain calculations.

        APDL Command: TREF

        Parameters
        ----------
        tref
            Reference temperature for thermal expansion.

        Notes
        -----
        Defines the reference temperature for the thermal strain calculations
        in structural analyses and explicit dynamic analyses.  Thermal strains
        are given by : ``α * (T-TREF)``, where α is the coefficient of thermal
        expansion (for more on this see the Mechanical APDL Theory Reference).
        Input the strain via ALPX, ALPY, ALPZ (the secant or mean coefficient
        value), or CTEX, CTEY, CTEZ (the instantaneous coefficient value), or
        the thermal strain value (THSX, THSY, THSZ). T is the element
        temperature. If α is temperature-dependent, TREF should be in the range
        of temperatures you define using the MPTEMP command.

        Reference temperatures may also be input per material by specifying a
        value on the MP material property command:

        MP,REFT,MAT,C0.

        Only a constant (non-temperature-dependent) value is valid. The value
        input on the TREF command applies to all materials not having a
        specified material property definition.

        To convert temperature-dependent secant coefficients of thermal
        expansion (SCTE) data (properties ALPX, ALPY, ALPZ) from the definition
        temperature to the reference temperature defined via a TREF (or
        MP,REFT) command, issue the MPAMOD command.

        This command is also valid in PREP7.
        """
        command = f"TREF,{tref}"
        return self.run(command, **kwargs)

    def tsres(self, array="", **kwargs):
        """Defines an array of key times at which the time-stepping strategy

        APDL Command: TSRES
        changes.

        Parameters
        ----------
        array
            Identifies an Nx1x1 array parameter containing the key times at
            which the heat transfer time-stepping strategy changes (the time
            step is reset to the initial time step based on DELTIM or NSUBST
            settings).  The array name must be enclosed by % signs (e.g.,
            %array%).  See ``*DIM`` for more information on array parameters.

        Notes
        -----
        Time values in the array parameter must be in ascending order and must
        not exceed the time at the end of the load step as defined on the TIME
        command. The time increment between time points in the array list must
        be larger than the initial time step defined on the DELTIM or NSUBST
        command.  Time values must also fall between the beginning and ending
        time values of the load step.  For multiple load step problems, you
        must either change the parameter values to fall between the beginning
        and ending time values of the load step or reissue the command with a
        new array parameter.  To clear the array parameter specification, issue
        TSRES,ERASE.  Results can be output at the requested time points if the
        array or time values in the array are also specified in the OUTRES
        command using FREQ=%array%.  Use this command to reset the time-
        stepping strategy within a load step. You may need to reset the time-
        stepping strategy when using tabular time-varying boundary conditions.

        See Steady-State Thermal Analysis of the Thermal Analysis Guide  for
        more information on applying boundary conditions via tabular input.
        See Transient Thermal Analysis of the Thermal Analysis Guide for more
        information on defining the key time array.
        """
        command = f"TSRES,{array}"
        return self.run(command, **kwargs)

    def upcoord(self, factor="", key="", **kwargs):
        """Modifies the coordinates of the active set of nodes, based on the

        APDL Command: UPCOORD
        current displacements.

        Parameters
        ----------
        factor
            Scale factor for displacements being added to nodal coordinates.
            If FACTOR = 1.0, the full displacement value will be added to each
            node, 0.5, half the displacement value will be added, etc.  If
            FACTOR = -1, the full displacement value will be subtracted from
            each node, etc.

        key
            Key for zeroing displacements in the database:

            OFF - Do not zero the displacements (default).

            ON - Zero the displacements.

        Notes
        -----
        The UPCOORD command uses displacements stored in the ANSYS database,
        and not those contained within the results file, Jobname.RST.  Nodal
        coordinates are updated each time the command is issued.  After
        updating, both the nodal displacements and rotations are set to zero if
        Key = ON.

        For structural solutions with an updated mesh, unless the coefficient
        matrix is otherwise reformed (e.g., a new analysis or NLGEOM,ON) it
        should first be reformed by issuing a KUSE,-1 command.

        UPCOORD should not be issued between load steps in structural analysis.

        For a multiphysics simulation where a CFD or electromagnetic field is
        being coupled to a structure undergoing large displacements, all (or a
        portion) of the surrounding field mesh may take part in the structural
        solution to "move" with the displacing structure.  You can use the
        UPCOORD command with a suitable FACTOR to update the coordinates of the
        nodes using the newly computed displacements.  The mesh will now
        conform with the displaced structure for subsequent field solutions.
        However, the mesh should always be restored to its original location by
        using an UPCOORD,FACTOR command before performing any subsequent
        structural solutions.  This is true for both repeated linear solutions,
        and for nonlinear restarts. (All saved displacements are relative to
        the original mesh location.)

        This command is not intended to replace either the large displacement
        or birth and death logic.

        This command is also valid in PREP7.
        """
        command = f"UPCOORD,{factor},{key}"
        return self.run(command, **kwargs)

    def usrcal(
        self,
        rnam1="",
        rnam2="",
        rnam3="",
        rnam4="",
        rnam5="",
        rnam6="",
        rnam7="",
        rnam8="",
        rnam9="",
        **kwargs,
    ):
        """Allows user-solution subroutines to be activated or deactivated.

        APDL Command: USRCAL

        Parameters
        ----------
        rnam1, rnam2, rnam3, . . . , rnam9
            User-defined solution subroutine names to be activated.  Up to nine
            may be defined on one command or multiple commands may be used.  If
            Rnam1 = ALL, activate all valid user subroutines.   If Rnam1 =
            NONE, deactivate all valid user subroutines.  All characters are
            required:

            USREFL - Allows user defined scalar field (body force) loads.

            USERCV - Allows user defined convection (surface) loads.

            USERPR - Allows user defined pressure (surface) loads.

            USERFX - Allows user-defined heat flux (surface) loads.

            USERCH - Allows user-defined charge density (surface) loads.

            USERFD - Computes the complex load vector for the frequency domain logic.

            USEROU - Allows user supplied element output.

            USERMC - Allows user control of the hygrothermal growth).

            USOLBEG - Allows user access before each solution.

            ULDBEG - Allows user access before each load step.

            USSBEG - Allows user access before each substep.

            UITBEG - Allows user access before each equilibrium iteration.

            UITFIN - Allows user access after each equilibrium iteration.

            USSFIN - Allows user access after each substep.

            ULDFIN - Allows user access after each load step.

            USOLFIN - Allows user access after each solution.

            UANBEG - Allows user access at start of run.

            UANFIN - Allows user access at end of run.

            UELMATX - Allows user access to element matrices and load vectors.

            UTIMEINC - Allows a user-defined time step, overriding the program-determined time step.

            UCNVRG - Allows user-defined convergence checking, overriding the program-determined
                     convergence.

        Notes
        -----
        Allows certain user-solution subroutines to be activated or deactivated
        (system-dependent).  This command only affects the subroutines named.
        Other user subroutines (such as user elements, user creep, etc.) have
        their own activation controls described with the feature.

        The routines are commented and should be listed after performing a
        custom installation from the distribution media for more details.  See
        also the Advanced Analysis Guide for a general description of user-
        programmable features.

        Users must have system permission, system access, and knowledge to
        write, compile, and link the appropriate subroutines into the program
        at the site where it is to be run.  All routines should be written in
        FORTRAN. (For more information on FORTRAN compilers please refer to
        either the ANSYS, Inc. Windows Installation Guide or the ANSYS, Inc.
        Linux Installation Guide for details specific to your platform or
        operating system.) Issue USRCAL,STAT to list the status of these user
        subroutines.  Since a user-programmed subroutine is a nonstandard use
        of the program, the verification of any ANSYS run incorporating these
        commands is entirely up to the user.  In any contact with ANSYS
        customer support regarding the performance of a custom version of the
        ANSYS program, you should explicitly state that a user programmable
        feature has been used.

        This command is also valid in PREP7.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"USRCAL,{rnam1},{rnam2},{rnam3},{rnam4},{rnam5},{rnam6},{rnam7},{rnam8},{rnam9}"
        return self.run(command, **kwargs)

    def wrfull(self, ldstep="", **kwargs):
        """Stops solution after assembling global matrices.

        APDL Command: WRFULL

        Parameters
        ----------
        ldstep
            Specify action to take:

            OFF or 0 - Turn off feature (default)

            N - Turn on feature and set it to stop after assembling the global matrices and
                writing the .FULL file for load step N.

        Notes
        -----
        This command is used in conjunction with the SOLVE command to generate
        the assembled matrix file (.FULL file) only.  The element matrices are
        assembled into the relevant global matrices for the particular analysis
        being performed and the .FULL file is written.  Equation solution and
        the output of data to the results file are skipped.  To dump the
        matrices written on the .FULL file into Harwell-Boeing format, use the
        HBMAT command in /AUX2. To copy the matrices to a postscript format
        that can be viewed graphically, use the PSMAT command.

        To use the LSSOLVE macro with this command, you may need to modify the
        LSSOLVE macro to properly stop at the load step of interest.

        This command only valid for linear static, full harmonic, and full
        transient analyses when the sparse direct solver is selected.  This
        command is also valid for buckling or modal analyses with any mode
        extraction method.  This command is not valid for nonlinear analyses.
        It is not supported in a linear perturbation analysis.

        In general, the assembled matrix file .FULL contains stiffness, mass,
        and damping matrices. However, the availability of the matrices depends
        on the analysis type chosen when the file is written.
        """
        command = f"WRFULL,{ldstep}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for common class between Archive, and result mesh."""
from ansys.mapdl.reader import _reader, _relaxmidside
from ansys.mapdl.reader.elements import ETYPE_MAP
from ansys.mapdl.reader.misc import unique_rows
import numpy as np
import pyvista as pv

INVALID_ALLOWABLE_TYPES = TypeError(
    "`allowable_types` must be an array " "of ANSYS element types from 1 and 300"
)

# map MESH200 elements to a pymapdl_reader/VTK element type (see elements.py)
MESH200_MAP = {
    0: 2,  # line
    1: 2,  # line
    2: 2,  # line
    3: 2,  # line
    4: 3,  # triangle
    5: 3,  # triangle
    6: 3,  # quadrilateral
    7: 3,  # quadrilateral
    8: 5,  # tetrahedron with 4 nodes
    9: 5,  # tetrahedron with 10 nodes
    10: 4,  # hex with 8 nodes
    11: 4,
}  # hex with 8 nodes

SHAPE_MAP = {  # from ELIST definition
    0: "",
    1: "LINE",
    2: "PARA",
    3: "ARC ",
    4: "CARC",
    5: "",
    6: "TRIA",
    7: "QUAD",
    8: "TRI6",
    9: "QUA8",
    10: "POIN",  # codespell-ignore
    11: "CIRC",
    12: "",
    13: "",
    14: "CYLI",
    15: "CONE",
    16: "SPHE",
    17: "",
    18: "",
    19: "PILO",
}
# element type to VTK conversion function call map
# 0: skip
# 1: Point
# 2: Line (linear or quadratic)
# 3: Shell
# 4: 3D Solid (Hexahedral, wedge, pyramid, tetrahedral)
# 5: Tetrahedral
# 6: Line (always linear)
TARGE170_MAP = {
    "TRI": 3,  # 3-Node Triangle
    "QUAD": 3,  # 4-Node Quadrilateral
    "CYLI": 0,  # Not supported (NS)  # Cylinder
    "CONE": 0,  # NS  # Cone
    "TRI6": 3,  # 6-Node triangle
    "SPHE": 0,  # NS  # Sphere
    "PILO": 1,  # Pilot Node
    "QUAD8": 3,  # 8-Node Quadrilateral
    "LINE": 2,  # Line
    "PARA": 2,  # Parabola
    "POINT": 1,  # Point
}


def _parse_vtk(
    mesh,
    allowable_types=None,
    force_linear=False,
    null_unallowed=False,
    fix_midside=True,
    additional_checking=False,
):
    """Convert raw ANSYS nodes and elements to a VTK UnstructuredGrid

    Parameters
    ----------
    fix_midside : bool, optional
        Adds additional midside nodes when ``True``.  When
        ``False``, missing ANSYS cells will simply point to the
        first node.

    """
    if not mesh._has_nodes or not mesh._has_elements:
        # warnings.warn('Missing nodes or elements.  Unable to parse to vtk')
        return pv.UnstructuredGrid()

    etype_map = ETYPE_MAP
    if allowable_types is not None:
        try:
            allowable_types = np.asarray(allowable_types)
        except:
            raise INVALID_ALLOWABLE_TYPES

        if not issubclass(allowable_types.dtype.type, np.integer):
            raise TypeError("Element types must be an integer array-like")

        if allowable_types.min() < 1 or allowable_types.max() > 300:
            raise INVALID_ALLOWABLE_TYPES

        etype_map = np.zeros_like(ETYPE_MAP)
        etype_map[allowable_types] = ETYPE_MAP[allowable_types]

    # ANSYS element type to VTK map
    type_ref = np.empty(2 << 16, np.int32)  # 131072
    type_ref[mesh._ekey[:, 0]] = etype_map[mesh._ekey[:, 1]]

    if allowable_types is None or 200 in allowable_types:
        for etype_ind, etype in mesh._ekey:
            # MESH200
            if etype == 200 and etype_ind in mesh.key_option:
                # keyoption 1 contains various cell types
                # map them to the corresponding type (see elements.py)
                mapped = MESH200_MAP[mesh.key_option[etype_ind][0][1]]
                type_ref[etype_ind] = mapped

            # TARGE170 specifics
            if etype == 170:
                # edge case where missing element within the tshape_key
                if etype_ind not in mesh.tshape_key:  # pragma: no cover
                    continue
                tshape_num = mesh.tshape_key[etype_ind]
                if tshape_num >= 19:  # weird bug when 'PILO' can be 99 instead of 19.
                    tshape_num = 19
                tshape_label = SHAPE_MAP[tshape_num]
                type_ref[etype_ind] = TARGE170_MAP.get(tshape_label, 0)

    nodes, angles, nnum = mesh.nodes, mesh.node_angles, mesh.nnum

    offset, celltypes, cells = _reader.ans_vtk_convert(
        mesh._elem, mesh._elem_off, type_ref, nnum, True
    )  # for reset_midside

    # fix missing midside
    if np.any(cells == -1):
        if fix_midside:
            nodes, angles, nnum = fix_missing_midside(
                cells, nodes, celltypes, offset, angles, nnum
            )
        else:
            cells[cells == -1] = 0

    if additional_checking:
        cells[cells < 0] = 0
        # cells[cells >= nodes.shape[0]] = 0  # fails when n_nodes < 20

    grid = pv.UnstructuredGrid(cells, celltypes, nodes, deep=True)

    # Store original ANSYS element and node information
    try:
        grid.point_data["ansys_node_num"] = nnum

    except ValueError:
        grid.point_data["ansys_node_num"] = (
            mesh._mapdl.nlist(kinternal="internal").to_array()[:, 0].astype(np.int32)
        )

    grid.cell_data["ansys_elem_num"] = mesh.enum
    grid.cell_data["ansys_real_constant"] = mesh.elem_real_constant
    grid.cell_data["ansys_material_type"] = mesh.material_type
    grid.cell_data["ansys_etype"] = mesh._ans_etype
    grid.cell_data["ansys_elem_type_num"] = mesh.etype

    # store node angles
    if angles is not None:
        if angles.shape[1] == 3:
            grid.point_data["angles"] = angles

    if not null_unallowed:
        grid = grid.extract_cells(grid.celltypes != 0)

    if force_linear:
        # only run if the grid has points or cells
        if grid.n_points:
            grid = grid.linear_copy()

    # map over element types
    # Add tracker for original node numbering
    ind = np.arange(grid.n_points)
    grid.point_data["origid"] = ind
    grid.point_data["VTKorigID"] = ind
    return grid


def fix_missing_midside(cells, nodes, celltypes, offset, angles, nnum):
    """Adds missing midside nodes to cells.

    ANSYS sometimes does not add midside nodes, and this is denoted in
    the element array with a ``0``.  When translated to VTK, this is
    saved as a ``-1``.  If this is not corrected, VTK will segfault.

    This function creates missing midside nodes for the quadratic
    elements.
    """
    # Check for missing midside nodes
    mask = cells == -1
    nnodes = nodes.shape[0]

    nextra = mask.sum()
    cells[mask] = np.arange(nnodes, nnodes + nextra)

    nodes_new = np.empty((nnodes + nextra, 3))
    nodes_new[:nnodes] = nodes
    nodes_new[nnodes:] = 0  # otherwise, segfault disaster

    # Set new midside nodes directly between their edge nodes
    temp_nodes = nodes_new.copy()
    _relaxmidside.reset_midside(cells, celltypes, offset, temp_nodes)

    # merge midside nodes
    unique_nodes, idx_a, idx_b = unique_rows(temp_nodes[nnodes:])

    # rewrite node numbers
    cells[mask] = (idx_b + nnodes).ravel()
    nextra = idx_a.shape[0]  # extra unique nodes
    nodes_new = nodes_new[: nnodes + nextra]
    nodes_new[nnodes:] = unique_nodes

    if angles is not None:
        new_angles = np.empty((nnodes + nextra, 3))
        new_angles[:nnodes] = angles
        new_angles[nnodes:] = 0
    else:
        new_angles = None

    # Add extra node numbers
    nnum_new = np.empty(nnodes + nextra)
    nnum_new[:nnodes] = nnum
    nnum_new[nnodes:] = -1
    return nodes_new, new_angles, nnum_new



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""These commands are used to parse responses from MAPDL"""
import re
from typing import Optional

NUMERIC_CONST_PATTERN = r"""
[-+]? # optional sign
(?:
(?: \d* \. \d+ ) # .1 .12 .123 etc 9.1 etc 98.1 etc
|
(?: \d+ \.? ) # 1. 12. 123. etc 1 12 123 etc
)
# followed by optional exponent part if desired
(?: [Ee] [+-]? \d+ ) ?
"""


NUM_PATTERN = re.compile(NUMERIC_CONST_PATTERN, re.VERBOSE)


def parse_kdist(msg):
    """Parse the keypoint value from a keypoint message"""
    finds = re.findall(NUM_PATTERN, msg)[-4:]
    if len(finds) == 4:
        return [float(val) for val in finds]


def parse_et(msg: Optional[str]) -> Optional[int]:
    """Parse local element type number definition message
    and return element type number.
    """
    if msg:
        res = re.search(r"(ELEMENT TYPE\s*)([0-9]+)", msg)
        if res is not None:
            return int(res.group(2))


def parse_e(msg: Optional[str]) -> Optional[int]:
    """Parse create element message and return element number."""
    if msg:
        res = re.search(r"(ELEMENT\s*)([0-9]+)", msg)
        if res is not None:
            return int(res.group(2))


def parse_kpoint(msg):
    """Parse create keypoint message and return keypoint number."""
    if msg:
        res = re.search(r"kpoint=\s+(\d+)\s+", msg)
        if res is not None:
            return int(res.group(1))


def parse_output_areas(msg):
    """Parse create area message and return area number."""
    if msg:
        res = re.search(r"(OUTPUT AREAS =\s*)([0-9]+)", msg)
        if res is not None:
            return int(res.group(2))
        res = re.search(r"(OUTPUT AREA\(S\) =\s*)([0-9]+)", msg)
        if res is not None:
            return int(res.group(2))


def parse_a(msg):
    """Parse create area message and return area number."""
    if msg:
        res = re.search(r"(AREA NUMBER =\s*)([0-9]+)", msg)
        if res is not None:
            return int(res.group(2))


def parse_line_no(msg):
    """Parse create line message and return line number."""
    if msg:
        res = re.search(r"LINE NO[.]=\s+(\d+)", msg)
        if res is not None:
            return int(res.group(1))


def parse_line_nos(msg):
    if msg:
        matches = re.findall(r"LINE NO[.]=\s*(\d*)", msg)
        if matches:
            return [int(match) for match in matches]


def parse_v(msg):
    """Parse volume message and return volume number"""
    if msg:
        res = re.search(r"(VOLUME NUMBER =\s*)([0-9]+)", msg)
        if res is not None:
            return int(res.group(2))


def parse_output_volume_area(msg):
    """Parse create area message and return area or volume number"""
    if msg:
        res = re.search(r"OUTPUT (AREA|VOLUME|AREAS) =\s*([0-9]+)", msg)
        if res is not None:
            return int(res.group(2))


def parse_ndist(msg):
    """Parse the node value from a node message"""
    finds = re.findall(NUM_PATTERN, msg)[-4:]
    if len(finds) == 4:
        return [float(val) for val in finds]



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class UsePass:
    def dcvswp(
        self,
        option="",
        elem="",
        cnum="",
        vmax="",
        vinc1="",
        vinc2="",
        gap="",
        **kwargs,
    ):
        """Performs a DC voltage sweep on a ROM element.

        APDL Command: DCVSWP

        Parameters
        ----------
        option
            Sweep option:

            GV - Perform voltage sweep up to given voltage Vmax.

        elem
            Element number of the ROM element for the ROM use pass analysis.

        cnum
            Number of sweep conductor.

        vmax
            Maximum voltage. For the PI option, this voltage should be below
            the pull-in voltage value.

        vinc1
            Voltage increment for Vmax (default = Vmax/20).

        vinc2
            Voltage increment for pull-in voltage (default = 1).

        gap
            Gap elements option:

            0 - Create gap elements (COMBIN40) (default).

        Notes
        -----
        Vinc1 is used to ramp the sweep conductor voltage from 0 to Vmax. Vinc2
        is used to increase the sweep conductor voltage from Vmax to the pull-
        in value if the PI sweep option is used.

        Because ramping the voltage may lead to the unstable region of an
        electromechanical system, DCVSWP might not converge when the sweep
        conductor voltage approaches the pull-in value. To avoid non-converged
        solutions, you should use the gap option to create a set of spring-gap
        elements (COMBIN40). By default, DCVSWP creates two spring-gap elements
        with opposite orientations for each active modal displacement DOF of
        the ROM element. The gap size is set to the maximum absolute values of
        the deflection range for the corresponding mode, as calculated by
        RMMSELECT or modified  using the RMMRANGE command. The spring constants
        are set to 1.E5 for all the COMBIN40 elements. Along with the spring-
        gap elements, DCVSWP creates a set of constraint equations relating the
        ROM element modal displacements DOF (EMF) and the displacement DOF (UX)
        of the gap elements. Constraining the modal displacements using the
        spring-gap elements allows DCVSWP to converge in the pull-in range. The
        DCVSWP macro has a limit of 900 equilibrium iterations. If this limit
        is not sufficient to reach convergence, try the advanced techniques
        given in Overcoming Convergence Problems in the Structural Analysis
        Guide. For more information on gap elements, see Using Gap Elements
        with ROM144 in the Coupled-Field Analysis Guide.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"DCVSWP,{option},{elem},{cnum},{vmax},{vinc1},{vinc2},{gap}"
        return self.run(command, **kwargs)

    def rmlvscale(
        self,
        nload="",
        fact1="",
        fact2="",
        fact3="",
        fact4="",
        fact5="",
        **kwargs,
    ):
        """Defines element load vector scaling for a ROM use pass.

        APDL Command: RMLVSCALE

        Parameters
        ----------
        nload
            Total number of load cases to be considered within a ROM use pass.
            If Nload = "DELETE", all defined load vectors are deleted.

        fact1, fact2, fact3, . . . , fact5
            Scale factors applied to load vectors (maximum 5). Defaults to 0.

        Notes
        -----
        Specifies the element load scale factor applied to a ROM analysis use
        pass.  Element load vectors are extracted from a Static Analysis using
        the RMNDISP command. Up to 5 element load vectors may be scaled and
        applied to a ROM use pass.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMLVSCALE,{nload},{fact1},{fact2},{fact3},{fact4},{fact5}"
        return self.run(command, **kwargs)

    def rmuse(self, option="", usefil="", **kwargs):
        """Activates ROM use pass for ROM elements.

        APDL Command: RMUSE

        Parameters
        ----------
        option
            Type of data to be plotted. Valid types are:

            1 or "ON" - Activates ROM use pass.

        usefil
            Name of the reduced displacement file (.rdsp) created by the ROM
            Use Pass (required field only for the Expansion Pass).

        Notes
        -----
        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMUSE,{option},{usefil}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class FeSurfaceLoads:
    def sf(self, nlist="", lab="", value="", value2="", **kwargs):
        """Specifies surface loads on nodes.

        APDL Command: SF

        Parameters
        ----------
        nlist
            Nodes defining the surface upon which the load is to be applied.
            Use the label ALL or P, or a component name.  If ALL, all selected
            nodes [NSEL] are used (default).  If P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).

        lab
            Valid surface load label. Load labels are listed under "Surface
            Loads" in the input table for each element type in the Element
            Reference.

        value
            Surface load value or table name reference for specifying tabular
            boundary conditions.

        value2
            Second surface load value (if any).

        Notes
        -----
        Individual nodes may not be entered for this command.  The node list is
        to identify a surface and the Nlist field must contain a sufficient
        number of nodes to define an element surface.  The loads are internally
        stored on element faces defined by the specified nodes.  All nodes on
        an element face (including midside nodes, if any) must be specified for
        the face to be used, and the element must be selected.

        If all nodes defining a face are shared by an adjacent face of another
        selected element, the face is not free and will not have a load
        applied.  If more than one element can share the same nodes (for
        example, a surface element attached to a solid element), select the
        desired element type before issuing the SF command. The SF command
        applies only to area and volume elements.

        For shell elements, if the specified nodes include face one (which is
        usually the bottom face) along with other faces (such as edges), only
        face one is used.  Where faces cannot be uniquely determined from the
        nodes, or where the face does not fully describe the load application,
        use the SFE command.  A load key of 1 (which is typically the first
        loading condition on the first face) is used if the face determination
        is not unique.  A uniform load value is applied over the element face.
        """
        command = f"SF,{nlist},{lab},{value},{value2}"
        return self.run(command, **kwargs)

    def sfbeam(
        self,
        elem="",
        lkey="",
        lab="",
        vali="",
        valj="",
        val2i="",
        val2j="",
        ioffst="",
        joffst="",
        lenrat="",
        **kwargs,
    ):
        """Specifies surface loads on beam and pipe elements.

        APDL Command: SFBEAM

        Parameters
        ----------
        elem
            Element to which surface load is applied.  If ALL, apply load to
            all selected beam elements (ESEL).  If Elem = P, graphical picking
            is enabled and all remaining command fields are ignored (valid only
            in the GUI).  A component name may be substituted in Elem.

        lkey
            Load key associated with surface load (defaults to 1).  Load keys
            (1, 2, 3, etc.) are listed under "Surface Loads" in the input table
            for each element type in the Element Reference.  For beam and some
            pipe elements, the load key defines the load orientation.

        lab
            Valid surface load label.  Load labels are listed under "Surface
            Loads" in the input table for each element type in the Element
            Reference.  Structural labels:  PRES (pressure).

        vali, valj
            Surface load values at nodes I and J.  If VALJ is blank, it
            defaults to VALI.  If VALJ is zero, a zero is used.

        val2i, val2j
            Second surface load values at nodes I and J.  Currently not used.

        ioffst, joffst
            Offset distance from node I (toward node J) where VALI is applied,
            and offset distance from node J (toward node I) where VALJ is
            applied, respectively.

        lenrat
            Offset distance flag:

            0  - Offset is in terms of length units (default).

            1  - Offset is in terms of a length ratio (0.0 to 1.0).

        Notes
        -----
        Specifies surface loads on the selected beam elements. Distributed
        loads are applied on a force-per-length basis (that is, the width of
        the underlying element is not considered). To list and delete surface
        loads applied with this command, use the SFELIST and SFEDELE commands,
        respectively.

        If no offset values (IOFFSET and JOFFSET) are specified, the load is
        applied over the full element length. Values may also be input as
        length fractions, depending on the LENRAT setting. For example,
        assuming a line length of 5.0, an IOFFST of 2.0 with LENRAT = 0 or an
        IOFFST of 0.4 with LENRAT = 1 represent the same point.  If JOFFST =
        -1, VALI is assumed to be a point load at the location specified via
        IOFFST, and VALJ is ignored. (IOFFSET cannot be equal to -1.) The
        offset values are stepped even if you issue a KBC,0 command.

        Offsets are only available for element types BEAM188 and PIPE288 if
        using the cubic shape function (KEYOPT(3) = 3) for those element types.

        To accumulate (add) surface loads applied with this command, use the
        SFCUM,,ADD command. Use the same offset values used on the previous
        SFBEAM command (for a given element face); otherwise, the loads do not
        accumulate. If no offsets are specified, the command applies the
        previous offset values.

        This command is also valid in PREP7.
        """
        command = f"SFBEAM,{elem},{lkey},{lab},{vali},{valj},{val2i},{val2j},{ioffst},{joffst},{lenrat}"
        return self.run(command, **kwargs)

    def sfcum(self, lab="", oper="", fact="", fact2="", **kwargs):
        """Specifies that surface loads are to be accumulated.

        APDL Command: SFCUM

        Parameters
        ----------
        lab
            Valid surface load label.  If ALL, use all appropriate labels.

        oper
            Accumulation key:

            REPL - Subsequent values replace the previous values (default).

            ADD - Subsequent values are added to the previous values.

            IGNO - Subsequent values are ignored.

        fact
            Scale factor for the first surface load value. A (blank) or '0'
            entry defaults to 1.0.

        fact2
            Scale factor for the second surface load value. A (blank) or '0'
            entry defaults to 1.0.

        Notes
        -----
        Allows repeated surface loads (pressure, convection, etc.) to be
        replaced, added, or ignored.  Surface loads are applied with the SF,
        SFE, and SFBEAM commands.  Issue the SFELIST command to list the
        surface loads.  The operations occur when the next surface load
        specifications are defined.  For example, issuing the SF command with a
        pressure value of 25 after a previous SF command with a pressure value
        of 20 causes the current value of that pressure to be 45 with the add
        operation, 25 with the replace operation, or 20 with the ignore
        operation.  All new pressures applied with SF after the ignore
        operation will be ignored, even if no current pressure exists on that
        surface.

        Scale factors are also available to multiply the next value before the
        add or replace operation.  A scale factor of 2.0 with the previous
        "add" example results in a pressure of 70.  Scale factors are applied
        even if no previous values exist.  Issue SFCUM,STAT to show the current
        label, operation, and scale factors.  Solid model boundary conditions
        are not affected by this command, but boundary conditions on the FE
        model are affected.

        Note:: : The FE boundary conditions may still be overwritten by
        existing solid model boundary conditions if a subsequent boundary
        condition transfer occurs.

        SFCUM does not work for tabular boundary conditions.

        This command is also valid in PREP7.
        """
        command = f"SFCUM,{lab},{oper},{fact},{fact2}"
        return self.run(command, **kwargs)

    def sfdele(self, nlist="", lab="", **kwargs):
        """Deletes surface loads.

        APDL Command: SFDELE

        Parameters
        ----------
        nlist
            Label defining where to find the list of nodes:

            ALL - Use all selected nodes [NSEL].  If P, use graphical picking in GUI.  A
                  component label may be substituted for Nlist.

        lab
            Valid surface load label.  If ALL, use all appropriate labels.  See
            the SF command for labels.

        Notes
        -----
        Deletes surface loads as applied with the SF command.  Loads are
        deleted only for the specified nodes on external faces of selected area
        and volume elements.  For shell elements, if the specified nodes
        include face one (which is usually the bottom face) along with other
        faces (such as edges), only the loads on face one will be deleted.  The
        element faces are determined from the list of selected nodes as
        described for the SF command.  See the SFEDELE command for deleting
        loads explicitly by element faces.

        This command is also valid in PREP7.
        """
        command = f"SFDELE,{nlist},{lab}"
        return self.run(command, **kwargs)

    def sfe(
        self,
        elem="",
        lkey="",
        lab="",
        kval="",
        val1="",
        val2="",
        val3="",
        val4="",
        **kwargs,
    ):
        """Specifies surface loads on elements.

        APDL Command: SFE

        Parameters
        ----------
        elem
            Element to which surface load applies.  If ALL, apply load to all
            selected elements [ESEL].  If Elem = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).  A component name may be substituted for Elem.

        lkey
            Load key or face number associated with surface load (defaults to
            1).  Load keys (1,2,3, etc.) are listed under "Surface Loads" in
            the input data table for each element type in the Element
            Reference.

        lab
            Valid surface load label.  Load labels are listed under "Surface
            Loads" in the input table for each element type in the Element
            Reference.

        kval
            Value key.  If Lab = PRES:

            0 or 1 - VAL1 through VAL4 are used as real components of pressures.

            2 - VAL1 through VAL4 are used as imaginary components of pressures.

        val1
            First surface load value (typically at the first node of the face)
            or the name of a table for specifying tabular boundary conditions.
            Face nodes are listed in the order given for "Surface Loads" in the
            input data table for each element type in the Element Reference.
            For example, for SOLID185, the item 1-JILK associates LKEY = 1
            (face 1) with nodes J, I, L, and K.  Surface load value VAL1 then
            applies to node J of face 1.  To specify a table, enclose the table
            name in percent signs (%), e.g., %tabname%.  Use the ``*DIM`` command
            to define a table.  Only one table can be specified, and it must be
            specified in the VAL1 position; tables specified in the VAL2, VAL3,
            or VAL4 positions will be ignored. VAL2 applies to node I, etc.
        """
        command = f"SFE,{elem},{lkey},{lab},{kval},{val1},{val2},{val3},{val4}"
        return self.run(command, **kwargs)

    def sfedele(self, elem="", lkey="", lab="", **kwargs):
        """Deletes surface loads from elements.

        APDL Command: SFEDELE

        Parameters
        ----------
        elem
            Element to which surface load deletion applies.  If ALL, delete
            load from all selected elements [ESEL].  If ELEM = P, graphical
            picking is enabled and all remaining command fields are ignored
            (valid only in the GUI).  A component name may be substituted for
            ELEM.

        lkey
            Load key associated with surface load (defaults to 1).  If ALL,
            delete surface loads for all load keys.

        lab
            Valid surface load label.  If ALL, use all appropriate labels.  See
            the SFE command for labels.

        Notes
        -----
        Deletes surface loads from selected elements.  See the SFDELE command
        for an alternate surface load deletion capability based upon selected
        nodes.

        This command is also valid in PREP7.
        """
        command = f"SFEDELE,{elem},{lkey},{lab}"
        return self.run(command, **kwargs)

    def sfelist(self, elem="", lab="", **kwargs):
        """Lists the surface loads for elements.

        APDL Command: SFELIST

        Parameters
        ----------
        elem
            Element at which surface load is to be listed.  If ALL (or blank),
            list loads for all selected elements [ESEL].  If ELEM = P,
            graphical picking is enabled and all remaining command fields are
            ignored (valid only in the GUI).  A component name may be
            substituted for ELEM.

        lab
            Valid surface load label.  If ALL (or blank), use all appropriate
            labels.

        Notes
        -----
        The surface loads listed correspond to the current database values. The
        database is not updated for surface loads in POST1. Surface loads
        specified in tabular form, however, do list their values corresponding
        to the current results set in POST1.

        For SURF151 or SURF152 elements with an extra node for radiation and/or
        convection calculations (KEYOPT(5) = 1), the bulk temperature listed is
        the temperature of the extra node. If the thermal solution does not
        converge, the extra node temperature is not available for listing.

        Film effectiveness and free stream temperatures specified by the SFE
        command (Lab = CONV) can only be listed by this command. The command
        lists film coefficients and bulk temperatures first and then film
        effectiveness and free stream temperatures below those values.

        Distributed ANSYS Restriction: In Distributed ANSYS within the SOLUTION
        processor, SFELIST support is not available for elements SURF151 and
        SURF152 when surface loading is applied via extra nodes (KEYOPT(5 > 0).
        If the command is issued under these circumstances, the resulting
        surface loads shown are not reliable.

        This command is valid in any processor.
        """
        command = f"SFELIST,{elem},{lab}"
        return self.run(command, **kwargs)

    def sffun(self, lab="", par="", par2="", **kwargs):
        """Specifies a varying surface load.

        APDL Command: SFFUN

        Parameters
        ----------
        lab
            Valid surface load label.  Load labels are listed under "Surface
            Loads" in the input table for each element type in the Element
            Reference.  Issue SFFUN,STATUS to list current command settings.

        par
            Parameter containing list of surface load values.  If Lab = CONV,
            values are typically the film coefficients and Par2 values (below)
            are typically the bulk temperatures.

        par2
            Parameter containing list of second surface load values (if any).
            If Lab = CONV, the Par2 values are typically the bulk temperatures.
            Par2 is not used for other surface load labels.

        Notes
        -----
        Specifies a surface load "function" to be used when the SF or SFE
        command is issued.  The function is supplied through an array parameter
        vector which contains nodal surface load values.  Node numbers are
        implied from the sequential location in the array parameter.  For
        example, a value in location 11 applies to node 11.  The element faces
        are determined from the implied list of nodes when the SF or SFE
        command is issued.  Zero values should be supplied for nodes that have
        no load.  A tapered load value may be applied over the element face.
        These loads are in addition to any loads that are also specified with
        the SF or SFE commands.  Issue SFFUN (with blank remaining fields) to
        remove this specification.  Issue SFFUN,STATUS to list current
        settings.

        Starting array element numbers must be defined for each array parameter
        vector.  For example, SFFUN,CONV,A(1,1),A(1,2) reads the first and
        second columns of array A (starting with the first array element of
        each column) and associates the values with the nodes.  Operations
        continue on successive column array elements until the end of the
        column. Another example to show the order of the commands:

        SFFUN does not work for tabular boundary conditions.

        This command is also valid in PREP7.
        """
        command = f"SFFUN,{lab},{par},{par2}"
        return self.run(command, **kwargs)

    def sfgrad(self, lab="", slkcn="", sldir="", slzer="", slope="", **kwargs):
        """Specifies a gradient (slope) for surface loads.

        APDL Command: SFGRAD

        Parameters
        ----------
        lab
            Valid surface load label.  Load labels are listed under "Surface
            Loads" in the input table for each element type in the Element
            Reference.

        slkcn
            Reference number of slope coordinate system (used with Sldir and
            SLZER to determine COORD).  Defaults to 0 (the global Cartesian
            coordinate system).

        sldir
            Slope direction in coordinate system SLKCN:

            X - Slope is along X direction (default).  Interpreted as R direction for non-
                Cartesian coordinate systems.

            Y - Slope is along Y direction.  Interpreted as  θ direction for non-Cartesian
                coordinate systems.

            Z - Slope is along Z direction.  Interpreted as Φ direction for spherical or
                toroidal coordinate systems.

        slzer
            Coordinate location (degrees for angular input) where slope
            contribution is zero (CVALUE = VALUE).  Allows the slope
            contribution to be shifted along the slope direction.  For angular
            input, SLZER should be between ±180° if the singularity [CSCIR]  is
            at 180° and should be between 0° and 360° if the singularity is at
            0°.

        slope
            Slope value (load per unit length or per degree).

        Notes
        -----
        Specifies a gradient (slope) for surface loads.  All surface loads
        issued with the SF, SFE, SFL, or SFA commands while this specification
        is active will have this gradient applied (for complex pressures, only
        the real component will be affected; for convections, only the bulk
        temperature will be affected).  The load value, CVALUE, calculated at
        each node is:

        CVALUE = VALUE + (SLOPE X (COORD-SLZER))

        where VALUE is the load value specified on the subsequent SF, SFE, SFL,
        or SFA commands and COORD is the coordinate value (in the Sldir
        direction of coordinate system SLKCN) of the node.  Only one SFGRAD
        specification may be active at a time (repeated use of this command
        replaces the previous specification with the new specification).  Issue
        SFGRAD (with blank fields) to remove the specification.  Issue
        SFGRAD,STAT to show the current command status.  The SFGRAD
        specification (if active) is removed when the LSREAD (if any) command
        is issued.

        SFGRAD does not work for tabular boundary conditions.

        This command is also valid in PREP7.
        """
        command = f"SFGRAD,{lab},{slkcn},{sldir},{slzer},{slope}"
        return self.run(command, **kwargs)

    def sflist(self, node="", lab="", **kwargs):
        """Lists surface loads.

        APDL Command: SFLIST

        Parameters
        ----------
        node
            Node at which surface load is to be listed.  If ALL (or blank),
            list for all selected nodes [NSEL].  If NODE = P, graphical picking
            is enabled and all remaining command fields are ignored (valid only
            in the GUI).  A component name may be substituted for NODE.

        lab
            Valid surface load label.  If ALL (or blank), use all appropriate
            labels.

        Notes
        -----
        Lists the surface loads as applied with the SF command.  Loads are
        listed only for the specified nodes on external faces of selected area
        and volume elements.  Use SFELIST for line elements. The surface loads
        listed correspond to the current database values. The database is not
        updated for surface loads in POST1. Surface loads specified in tabular
        form, however, do list their values corresponding to the current
        results set in POST1.

        For SURF151 or SURF152 elements with an extra node for radiation and/or
        convection calculations (KEYOPT(5) = 1), the bulk temperature listed is
        the temperature of the extra node. If the thermal solution does not
        converge, the extra node temperature is not available for listing.

        This command is valid in any processor.
        """
        command = f"SFLIST,{node},{lab}"
        return self.run(command, **kwargs)

    def sfscale(self, lab="", fact="", fact2="", **kwargs):
        """Scales surface loads on elements.

        APDL Command: SFSCALE

        Parameters
        ----------
        lab
            Valid surface load label.  If ALL, use all appropriate labels.

        fact
            Scale factor for the first surface load value.  Zero (or blank)
            defaults  to 1.0.  Use a small number for a zero scale factor.

        fact2
            Scale factor for the second surface load value.  Zero (or blank)
            defaults  to 1.0.  Use a small number for a zero scale factor.

        Notes
        -----
        Scales surface loads (pressure, convection, etc.) in the database on
        the selected elements.  Surface loads are applied with the SF, SFE, or
        SFBEAM commands.  Issue the SFELIST command to list the surface loads.
        Solid model boundary conditions are not scaled by this command, but
        boundary conditions on the FE model are scaled.

        Note:: : Such scaled FE boundary conditions may still be overwritten by
        unscaled solid model boundary conditions if a subsequent boundary
        condition transfer occurs.

        SFSCALE does not work for tabular boundary conditions.

        This command is also valid in PREP7 and in the /MAP processor.
        """
        command = f"SFSCALE,{lab},{fact},{fact2}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Listing:
    def format(
        self,
        ndigit="",
        ftype="",
        nwidth="",
        dsignf="",
        line="",
        char="",
        **kwargs,
    ):
        """Specifies format controls for tables.

        APDL Command: /FORMAT

        Parameters
        ----------
        ndigit
            Number of digits (3 to 32) in first table column (usually the node
            or element number).  Initially defaults to 7.

        ftype
            FORTRAN format types (initially defaults to G):

            G - Gxx.yy.  xx and yy are described below.

            F - Fxx.yy

            E - Exx.yy

        nwidth
            Total width (9 to 32) of the field (the xx in Ftype).  Initially
            defaults to 12.

        dsignf
            Number of digits after the decimal point (yy in F or E format) or
            number of significant digits in G format.  Range is 2 to xx-7 for
            Ftype = G or E; and 0 to xx-4 for Ftype = F.  Initially defaults to
            5.

        line
            Number of lines (11 minimum) per page.  Defaults to ILINE or BLINE
            from the /PAGE command.

        char
            Number of characters (41 to 240, system-dependent) per line before
            wraparound.  Defaults to ICHAR or BCHAR from the /PAGE command.

        Notes
        -----
        Specifies various format controls for tables printed with the POST1
        PRNSOL, PRESOL, PRETAB, PRRSOL, PRPATH, and CYCCALC commands.  A blank
        (or out-of-range) field on the command retains the current setting.
        Issue /FORMAT,STAT to display the current settings.  Issue /FORMAT,DEFA
        to reestablish the initial default specifications.

        For the POST26 PRVAR command, the Ftype, NWIDTH, and DSIGNF fields
        control the time output format.

        This command is valid in any processor.
        """
        command = f"/FORMAT,{ndigit},{ftype},{nwidth},{dsignf},{line},{char}"
        return self.run(command, **kwargs)

    def header(
        self,
        header="",
        stitle="",
        idstmp="",
        notes="",
        colhed="",
        minmax="",
        **kwargs,
    ):
        """Sets page and table heading print controls.

        APDL Command: /HEADER

        Parameters
        ----------
        header
            ANSYS page header (system, date, time, version, copyright, title,
            etc.):

            ON - Turns this item on (default for batch mode; not available for interactive
                 mode).

            OFF - Turns this item off.

            (blank) - Retains the previous setting.

        stitle
            Subtitles (see /STITLE command):   ON, OFF, or (blank) (see above).

        idstmp
            Load step information (step number, substep number, time value):
            ON, OFF, or (blank) (see above).

        notes
            Information relative to particular table listings:   ON, OFF, or
            (blank) (see above).

        colhed
            Column header labels of table listings (currently only for single
            column tables):   ON, OFF, or (blank) (see above).

        minmax
            Minimum/maximum information or totals after table listings:   ON,
            OFF, or (blank) (see above).

        Notes
        -----
        Sets specifications on or off for page and table heading print controls
        associated with the POST1 PRNSOL, PRESOL, PRETAB, PRRSOL, and PRPATH
        commands.

        Note:: : If the printout caused a  top-of-form (page eject to top of
        next page), the top-of-form is also suppressed with the printout.
        Issue /HEADER,STAT to display the current settings.  Issue /HEADER,DEFA
        to reset the default specifications.

        This command is valid in any processor.
        """
        command = f"/HEADER,{header},{stitle},{idstmp},{notes},{colhed},{minmax}"
        return self.run(command, **kwargs)

    def irlist(self, **kwargs):
        """Prints inertia relief summary table.

        APDL Command: IRLIST

        Notes
        -----
        Prints the inertia relief summary data, including the mass summary
        table, the total load summary table, and the inertia relief summary
        table resulting from the inertia relief calculations. These
        calculations are performed in the solution phase [SOLVE] as specified
        by the IRLF command.

        Inertia relief output is stored in the database rather than in the
        results file (Jobname.RST). When you issue IRLIST, ANSYS pulls the
        information from the database, which contains the inertia relief output
        from the most recent solution [SOLVE].

        This command is valid in any processor.
        """
        command = f"IRLIST,"
        return self.run(command, **kwargs)

    def page(self, iline="", ichar="", bline="", bchar="", comma="", **kwargs):
        """Defines the printout and screen page size.

        APDL Command: /PAGE

        Parameters
        ----------
        iline
            Number of lines (11 minimum) per "page" or screen.  Defaults to 24.
            Applies to interactive non-GUI to the screen output only.

        ichar
            Number of characters (41 to 132) per line before wraparound.
            Defaults to 80.  Applies to interactive non-GUI to the screen
            output only.

        bline
            Number of lines (11 minimum) per page.  Defaults to 56.  Applies to
            batch mode [/BATCH], diverted [/OUTPUT], or interactive GUI [/MENU]
            output. If negative, no page headers are output.

        bchar
            Number of characters (41 to 240 (system dependent)) per line before
            wraparound.  Defaults to 132.  Applies to batch mode [/BATCH],
            diverted [/OUTPUT], or interactive GUI [/MENU] output.

        comma
            Input 1 to specify comma-separated output for node [NLIST] and
            element [ELIST] output.

        Notes
        -----
        Defines the printout page size for batch runs and the screen page size
        for interactive runs.  Applies to the POST1 PRNSOL, PRESOL, PRETAB,
        PRRSOL, and PRPATH commands.  See the /HEADER command for additional
        controls (page ejects, headers, etc.) that affect the amount of
        printout.  A blank (or out-of-range) value retains the previous
        setting.  Issue /PAGE,STAT to display the current settings.  Issue
        /PAGE,DEFA to reset the default specifications.

        This command is valid in any processor.
        """
        command = f"/PAGE,{iline},{ichar},{bline},{bchar},{comma}"
        return self.run(command, **kwargs)

    def prerr(self, **kwargs):
        """Prints SEPC and TEPC.

        APDL Command: PRERR

        Notes
        -----
        Prints the percent error in structural energy norm (SEPC) and the
        thermal energy norm percent error (TEPC).  Approximations of mesh
        discretization error associated with a solution are calculated for
        analyses having structural or thermal degrees of freedom.

        The structural approximation is based on the energy error (which is
        similar in concept to the strain energy) and represents the error
        associated with the discrepancy between the calculated stress field and
        the globally continuous stress field (see POST1 - Error Approximation
        Technique in the Mechanical APDL Theory Reference).  This discrepancy
        is due to the assumption in the elements that only the displacements
        are continuous at the nodes.  The stress field is calculated from the
        displacements and should also be continuous, but generally is not.

        Thermal analyses may use any solid and shell thermal element having
        only temperature degrees of freedom.  The thermal approximation is
        based on the total heat flow dissipation and represents the error
        associated with the discrepancy between the calculated nodal thermal
        flux within an element and a continuous global thermal flux.  This
        continuous thermal flux is calculated with the normal nodal averaging
        procedure.

        The volume (result label VOLU) is used to calculate the energy error
        per element (result label SERR for the structural energy error and TERR
        for the thermal energy error).  These energy errors, along with the
        appropriate energy, are then used to calculate the percent error in
        energy norm (SEPC for structural and TEPC for thermal). These
        percentages can be listed by the PRERR command, retrieved by the ``*GET``
        command (with labels SEPC and TEPC) for further calculations, and shown
        on the displacement display (PLDISP), as applicable.

        For structural analyses, the maximum absolute value of nodal stress
        variation of any stress component for any node of an element (result
        item SDSG) is also calculated.  Similarly, for thermal gradient
        components, TDSG is calculated.  Minimum and maximum result bounds
        considering the possible effect of discretization error will be shown
        on contour displays (PLNSOL).  For shell elements, the top surface
        location is used to produce a meaningful percentage value.  SERR, TERR,
        SEPC, TEPC, SDSG, and TDSG will be updated whenever the nodal stresses
        or fluxes are recalculated.

        If the energy error is a significant portion of the total energy, then
        the analysis should be repeated using a finer mesh to obtain a more
        accurate solution.  The energy error is relative from problem to
        problem but will converge to a zero energy error as the mesh is
        refined.  An automated adaptive meshing procedure using this energy
        error is described with the ADAPT macro.

        The following element- and material-type limitations apply:

        Valid with most 2-D solid, 3-D solid, axisymmetric solid, or 3-D shell
        elements.

        The following element types are not valid: SHELL28, SHELL41, and
        SOLID65.

        The model should have only structural or thermal degrees of freedom.

        The analysis must be linear (for both material and geometry).

        Multi-material (for example, composite) elements are not valid.

        Transition regions from one material to another are not valid (that is,
        the entire model should consist of one material).

        Anisotropic materials (TB,ANEL) are not considered.
        """
        command = f"PRERR,"
        return self.run(command, **kwargs)

    def priter(self, **kwargs):
        """Prints solution summary data.

        APDL Command: PRITER

        Notes
        -----
        Prints solution summary data (such as time step size, number of
        equilibrium iterations, convergence values, etc.) from a static or full
        transient analysis. All other analyses print zeros for the data.
        """
        command = f"PRITER,"
        return self.run(command, **kwargs)
